#!/usr/bin/env python
"""

# Spec Building

## Dev
### Frequent abbrs:
- rt: runtime (e.g. systemd, docker, k8s,)
- cls: A class in a spec python file, denoting a config entity
- fid: Full id of an item (path in the spec : ['Wifi', 'NB01', 'Redis']
"""
# stream of specs - from anywhere, we will run as daemon some day:
# before that its just a one item stream, with item pushed at proc stasrt:

import os
import subprocess as sp

import rx
from devapp.app import app
from devapp.spec import find_paths_in_fs_components
from devapp.spec.tools import FS, full_url_and_path
from devapp.tools import (
    dirname,
    download_file,
    exists,
    into,
    is_,
    sp_call,
    to_list,
    url_to_dir,
)
from devapp.utils.rx_tools import chain, do_if, threaded
from rx import operators as op
from rx.operators import filter

base_dirs = 'var', 'data', 'conf', 'log', 'build', 'etc'

# filesystem components as declared in specs:
fs_comps = {}

default_methods = {'git': 'mount', 'oci': 'overlay', 'conda_env': 'activate'}

pass_ = lambda a: a

# fmt:off
d_oci         = lambda r: r['env']['DA_DIR_OCI']   + '/' + r['path']
props         = lambda r: r['spec']['props']
frm_to        = lambda frm, to: {'from': frm, 'to': to}
d_repo        = lambda r: r['env']['DA_DIR_REPOS'] + '/' + r['path']
comp_name     = lambda p: p if is_(p, str) else p['name']
rfs_comps     = lambda r: [fs_comps[comp_name(p)] for p in filesystem(r)]
filesystem    = lambda r: props(r).get('filesystem')
d_exe_bare    = lambda env: env['DA_DIR_OCI'] + '/executables'

# fmt:on
def d_checkout(repo):
    r = repo
    d, p = '', r['path']
    if '/' in p:
        d, p = p.rsplit('/', 1)
        d = '@' + d.replace('/', sep_slash)
    res = r.get('checkout_dir') or (
        r['env']['DA_DIR_CHECKOUTS']
        + '/'
        + p
        + d
        + sep_branch
        + r['branch'].replace('/', sep_slash)
    )
    return res


# we want one flat dir of checkouts. need to name them somewhat sensibly
# they should show the branch they are based on
sep_branch = '#'
sep_slash = '^'

root_fs_mount_ph = '$fs'

# Phase 1: FETCH FS COMPONENTS------------------------------------------------:
# The First Phase: Walking all filesysem components, independent of classes


class FSComponents:
    """
    Namespace for Component Specific Methods
    We often use the word repo for component, for analogy with git
    """

    class Defaults:
        # ready to be adapted, e.g. by flags
        exe_fs_method = 'symlink'
        local_unknown_comptype = 'ext_dir'

    class host_system:
        """host packages"""

        def prepare(repo):
            repo['bare_have'] = True
            repo['skip_completion'] = True
            repo['skip_add_fs_presence_infos'] = True
            repo['packages'] = to_list(repo['packages'])
            # just something meaningful, packages are handled by the distri:
            repo['checkout_dir'] = '/usr/bin'
            return repo

        def get_bare(repo):
            return repo

        def _detect_inst_cmd():
            ps = 'dnf', 'yum', 'apt-get'
            for c in ps:
                cmd = getattr(app.sh, c, None)
                if cmd:
                    break
            if not cmd:
                app.die('Could not derive a package installer', tried=ps)

            return (c, 'install', '-y')

        @classmethod
        def into_fs(pkg, repo):
            inst_cmd = pkg._detect_inst_cmd()
            inst_cmd += tuple(repo['packages'])
            st = sp_call(*inst_cmd, as_root=True, get_all=True)
            if st['exit_status']:
                app.die('Package install failed', **st)
            return repo

    class git:
        def get_bare(repo):
            app.sh.git.clone('--bare', repo['url'], d_repo(repo))
            repo['bare_have'] = True
            return repo

        def checkout(repo):
            d = d_checkout(repo)
            app.sh.git.clone(d_repo(repo), d)
            app.sh.git.checkout(repo['branch'], _cwd=d)
            repo['checkout_have'] = True
            return repo

        def set_remotes(repo):
            d = d_checkout(repo)
            # have = app.sh.git.remotes('-v', _cwd=d)
            app.sh.git.remote.rename('origin', 'local', _cwd=d)
            app.sh.git.remote.add('origin', repo['url'], _cwd=d)
            return repo

        into_fs = [checkout, set_remotes]

    class oci:
        def prepare(repo):
            repo['branch'] = 'oci'
            repo['run_dir'] = root_fs_mount_ph  # will be replaced later
            # oci always are overlayed, no matter the run_dir:
            repo['fs_method'] = repo.get('fs_method') or default_methods['oci']
            return repo

        def pull_layers(repo):
            url = repo['url']
            # app.sh.dpull('-d', repo['bare_dir'], '--repo', url)
            err = sp.call(['dpull', '-d', repo['bare_dir'], '--repo', url])
            if err:
                app.die('could not pull layers', **repo)
            repo['bare_have'] = True
            return repo

        def merge_layers(repo):
            b, c = repo['bare_dir'], repo['checkout_dir']
            err = sp.call(['dmerge', '-d', b, '--method', 'tar', '-t', c])
            if err:
                app.die('could not merge layers', **repo)
            repo['checkout_have'] = True
            return repo

        get_bare = pull_layers
        into_fs = merge_layers

    class mount:
        def prepare(repo):
            repo['branch'], n, D = 'mount', repo['name'], repo['env']['DA_DIR']
            repo['base_name'] = repo['name']
            repo['fs_method'] = 'mount'  # checkout will be mounted to run
            repo['run_dir'] = repo['to']
            repo['checkout_dir'] = repo['bare_dir'] = repo['from']
            repo['bare_have'] = repo['checkout_have'] = True
            repo['skip_completion'] = True
            return repo

        def get_bare(repo):
            return repo

    class conda_env:
        def prepare(repo):
            repo['branch'], n, D = 'env', repo['name'], repo['env']['DA_DIR']
            repo['base_name'] = repo['name']
            repo['checkout_dir'] = repo['bare_dir'] = D + '/envs/' + n
            repo['run_dir'] = repo['checkout_dir']
            repo['bare_have'] = True
            repo['skip_completion'] = True
            repo['fs_method'] = 'pass'  # done via link to env's python (in build.sh)
            return repo

        def get_bare(repo):
            return repo

        def into_fs(repo):
            D = repo['env']['DA_DIR']
            cmds = [
                '-p',
                D + '/envs/%s' % repo['name'],
                '-y',
                *[l for l in to_list(repo['packages']) if is_(l, str)],
            ]
            for c in to_list(repo.get('add_channels')):
                cmds.insert(0, c)
                cmds.insert(0, '-c')

            app.info('conda create', args=' '.join(cmds))
            app.sh.conda.create(*cmds)
            l = [(l[0], to_list(l[1:])) for l in repo['packages'] if not is_(l, str)]
            icmd = app.sh.conda.install
            if l:
                for chan, ps in l:
                    app.info('conda.install', args=' '.join(cmds))
                    icmd('-c', chan, '-n', repo['name'], '-y', *ps)
            pips = repo.get('pips', ())
            if pips:
                m = {'DA_DIR': repo['env']['DA_DIR'], 'name': repo['name']}
                pc = '%(DA_DIR)s/envs/%(name)s/bin/pip' % m
                for pip in pips:
                    if sp.call([pc, 'install', pip]):
                        app.die('Pip failed', pip=pip)
            repo['checkout_have'] = True
            return repo

    class ext_dir:
        """Using anywhere on the filesystem"""

        @classmethod
        def get_bare(sl, repo):
            for d in 'bare_dir', 'checkout_dir', 'run_dir', 'path':
                repo[d] = repo['url'].replace('file://', '')
            return repo

    class exe:
        """Simple static binaries, gitlab runner, hashitools in go...
        We pull to bare, copy to checkout, mount or copy to run_dir
        """

        _dest_dir = '/bin/'

        @classmethod
        def fn(exe, where, repo):
            fn = repo[where] + exe._dest_dir + repo['name']
            if not exists(dirname(fn)):
                os.makedirs(dirname(fn))
            return fn

        def prepare(repo):
            repo['branch'], env = 'exe', repo['env']
            d = url_to_dir(repo['url'])
            for k, m in (('bare_dir', d_exe_bare),):
                repo[k] = '/'.join((m(env), d))
            repo['branch'] = 'exe'
            repo['fs_method'] = repo.get(
                'fs_method', FSComponents.Defaults.exe_fs_method
            )
            return repo

        @classmethod
        def get_bare(exe, repo):
            fn = exe.fn('bare_dir', repo)
            download_file(repo['url'], local_filename=fn)
            return repo

        @classmethod
        def into_fs(exe, repo):
            fn = exe.fn('checkout_dir', repo)
            app.sh.cp('-f', exe.fn('bare_dir', repo), fn)
            app.sh.chmod('+x', fn)
            return repo

    class lib(exe):
        _dest_dir = '/'


def fetch_fs_comps(env, processing=[]):
    """At spec loading we walked thru all classes and generated the list of
    fs components in use in the spec"""
    # todo: Node local only comps filter, either here or at spec walk.

    return chain(
        # sort just to get same download behaviour(order) everywhere:
        rx.from_(sorted(fs_comps.items())),
        filter(lambda kv: kv[0] not in processing),
        op.do_action(lambda kv: processing.append(kv[0])),  # don't do twice
        lambda kv: kv[1],  # we drop the name, was just to sort
        lambda repo, env=env: into(repo, 'env', env),
        do_if(_flow_by_type, which=_guess_component_type),
        # ldd, # uncomment to check produced infos per component
    )


# --------------------------------------------------- Helpers for the main flow
def _guess_component_type(repo_as_from_spec):
    """for group by"""
    repo = repo_as_from_spec
    if 'xhello' in str(repo):
        breakpoint()
    un, f = repo.get('url') or repo['name'], 'file://'
    # local?
    un, type = un.replace(f, ''), repo.get('type')
    # yes we allow even relative paths:
    if un.startswith('/') or un.startswith('./'):
        d = FSComponents.Defaults.local_unknown_comptype
        if not os.path.exists(un):
            app.die('Not found', url=un)
        if os.path.exists(un + '/.git') or un.endswith('.git'):
            repo['type'] = type = type or 'git'
        elif os.path.exists(un + '/.hg'):
            app.die('Mercurial not yet supported', d=un)
        elif not type:
            repo['type'] = type = type or d
    return (
        type
        if type
        else 'git'
        if (un.endswith('.git') or '.git#' in un)
        else 'oci'
        if len(un.split(':')) == 2
        else None
    )


def _flow_by_type(comp_type):
    """flows are so similar, we offer this for oci and git
    """
    cls = getattr(FSComponents, comp_type)
    ifs = getattr(cls, 'into_fs', pass_)
    into_fs = [ifs] if callable(ifs) else ifs
    return [
        lambda repo: into(repo, 'type', comp_type),
        getattr(cls, 'prepare', pass_),
        _complete_repo_infos,
        _add_fs_presence_infos,
        do_if(cls.get_bare, threaded(10), if_=lambda comp: not comp.get('bare_have'),),
        # do_if(*into_fs, if_=lambda comp: not comp.get('checkout_have')),
        do_if(*into_fs, if_=checkout_missing),
        _find_auto_env_paths,
        _drop_env,
    ]


def checkout_missing(comp):
    return not comp.get('checkout_have')


def _complete_repo_infos(repo):
    """Repo as derived from initial spec walk
    If in spec there was only string, then it has only 'name' set.
    """
    type, env, url = repo['type'], repo['env'], repo.get('url') or repo['name']
    if repo.get('skip_completion'):
        return repo
    app.debug('Repo type', type=type, url=url)
    repo['base_name'] = url.rsplit('/', 1)[-1].replace('.git', '')
    url += '#' + repo.get('branch', 'master')
    url, repo['branch'] = url.split('#')[:2]
    # d = url.split('/')
    # if len(d) > 1 and exists(d[1]):
    #    d = 'file://' + d
    repo['url'], repo['path'] = full_url_and_path(env, url, mode=repo['type'])
    repo['bare_dir'] = repo.get('bare_dir') or (
        d_repo(repo) if type == 'git' else d_oci(repo)
    )
    repo['checkout_dir'] = dc = d_checkout(repo)
    rd = repo.get('run_dir')
    if rd and rd != dc:
        # if there is a run_dir set, we have to mount over:
        rd = rd if rd[0] in ('/', '$') else ('%(DA_DIR)s/run/' % env + rd)
        repo['run_dir'] = rd
        repo['fs_method'] = repo.get('fs_method') or default_methods[type]
    else:
        # otherwise we'll just run from the checkout dir:
        # no mount required, then, for git
        repo['run_dir'] = dc
    return repo


def _add_fs_presence_infos(repo):
    if repo.get('skip_add_fs_presence_infos'):
        return repo
    d = repo['bare_dir']
    if exists(d):
        repo['bare_have'] = True
        repo['bare_is_up_to_date'] = 'maybe'
    # the co have check dir feature enables to detail when a checkout is considered present.
    # usefual for e.g. gitlab_runner/hugo, with bin and themes repos:
    d = repo.get('checkout_have_check_dir') or repo['checkout_dir']
    if exists(d):
        repo['checkout_have'] = True
        repo['checkout_is_up_to_date'] = 'maybe'
    return repo


def _find_auto_env_paths(repo):
    """A filesystem layer usually contains stuff processes want to use ;-)
    Find and define those here
    """
    cod = repo['checkout_dir']
    sds = [cod]  # searchdirs
    dd = cod + '/daemon'
    # if 'gbase' in str(repo):
    #    breakpoint()
    finders = find_paths_in_fs_components
    if exists(dd):
        [sds.insert(0, dd + '/' + d) for d in os.listdir(dd)]
    chain(
        rx.from_([(repo, cod, s) for s in sds]),
        finders.find_bin_paths,
        finders.find_py_paths,
    ).run()
    return repo


def _drop_env(repo):
    repo.pop('env')
    return repo


def check_mk_dirs(env):
    for k in ('DA_DIR_OCI', 'DA_DIR_CHECKOUTS'):
        if not k in env:
            env[k] = env['DA_DIR'] + '/' + k.rsplit('_', 1)[-1].lower()
    for k in ('DA_DIR_REPOS',):
        if not k in env:
            app.die('Require', env_key=k)
    dirs = ('DA_DIR_REPOS', 'DA_DIR_OCI', 'DA_DIR_CHECKOUTS')
    [app.sh.mkdir('-p', env[k]) for k in dirs if not exists(env[k])]


# Phase 2: Apply Per Class ----------------------------------------------------
# 'r' dicts are sent in per class of the spec:
def set_run_and_checkout_dirs(rc):
    r, comp = rc
    for d in 'run_dir', 'checkout_dir':
        r['env'][d] = comp[d]
    return rc


def set_mount(rc):
    r, comp = rc
    rd, cd = (comp['run_dir'], comp['checkout_dir'])
    if comp['type'] == 'exe':
        rd, cd = rd + '/' + comp['name'], cd + '/' + comp['name']
    if not FS.can_mount():
        app.warn('Require bindfs', frm=rd, to=cd)
    r['fs'][rd] = {'meth': 'mount', 'from': cd}
    return rc


def set_overlay(rc):
    r, comp = rc
    mp = comp['run_dir']
    have, cd = r['fs'].get(mp), comp['checkout_dir']
    if have:
        have['from'].append(cd)
    else:
        r['fs'][mp] = {'meth': 'overlay', 'from': [cd]}
    return rc


def add_framework_dirs(rc):
    r, comp = rc
    e, D = r['env'], r['env']['DA_DIR']
    r['fs']['$fs' + D] = {'meth': 'mount', 'from': D}
    bds = list(base_dirs)
    bds.append('repos')
    for d in bds:
        db = e.get('DA_DIR_%s' % d.upper())
        # outside one:
        if db and not db.startswith(D):
            r['fs']['$fs' + db] = {'meth': 'mount', 'from': db}
    return rc


fs_methods = {'mount': set_mount, 'overlay': [set_overlay, add_framework_dirs]}


def ldd(r):
    breakpoint()
    return r


def define_fs_stack(r):
    chain(
        rx.from_([(r, c) for c in r['fs_comps']]),
        set_run_and_checkout_dirs,
        do_if(fs_methods, which=lambda rc: rc[1].get('fs_method')),
    ).run()
    return r



#!/usr/bin/env python
# coding: utf-8
# Some systems do not accept shebang with args. Use python -Ss mdvl.py then.
"""
# Lightweight Simple Markdown Renderer for the Terminal

## Usage

    mdvl <markdown source | markdown file>
    cat <markdown file> | mdvl

## Config

```
%s
```

### Colors

```
%%s
```

## Debugging Parsing Errors

    export mdvl_debug=1

See also https://www.bing.com/search?q=github+mdvl
"""
__version__ = '2017.07.16.7'  # count up for new pip versions
__author__ = 'Gunther Klessinger'

import os
import re
from operator import setitem as set
from textwrap import fill

debug = os.environ.get('mdvl_debug')

# check environ for value and cast into bools if necessary:
_b = {'True': True, 'False': False}
env = lambda k, d=None: _b.get(k, os.environ.get(k, d))

# ----------------------------------------------------------------- Config Mgmt
class Cfg:
    """
    Base class for osenv and kw configurable instances.
    - We have defaults, overridable by environ keys, overridable by **kws
    - problem is that color codes should be givabable as ints - but on the
      terminal we usually have them as full ansi escape str.
      Thats why `get_val` is present and adapts for that in the Colors cls.
    """

    _parms = None  # our (relevant) keys and values

    def setup(self, kw):
        """ find all our key value defaults and override with env and **kw"""
        self._parms = []
        kv = [(k, getattr(self, k)) for k in dir(self) if not k.startswith('_')]
        self._parms = [(k, v) for k, v in kv if not hasattr(v, '__code__')]
        [setattr(self, k, self.get_val(k, v, kw)) for k, v in self._parms]

    def get_val(self, k, dflt, kw):
        try:
            return type(dflt)(kw.get(k, env(k, dflt)))
        except Exception as ex:
            raise Exception(
                'Could not derive type for %s %s %s %s' % (k, dflt, kw, env(k, dflt))
            )


class Colors(Cfg):
    """
    Color namespae with efault color scheme (greenish).
    The 'C' in the main method.
    x_ -> x with ansi escapes in __init__, with env precedence
    """

    O = '\x1B[0m'
    GRAY = 240
    CODE = 5
    L = env('L', 5)
    H1 = env('I', 11)
    H2 = env('G', 2)
    H3 = env('M', 1)
    H4 = env('CODE', 6)
    emph = env('I', 11)
    ital = env('M', 5)

    def get_val(self, k, dflt, kw):
        # see Cfg for expl.
        v = kw.get(k, env(k, dflt))
        v = str(v)
        # if k == 'H1':
        #    breakpoint()
        e = 1 if k[0] == 'H' else 0
        if v.isdigit():
            v = '\x1B[%s;38;5;%sm' % (e, v)
        else:
            v = '\x1b[' + v.split('[', 1)[1]
        return v

    def H(s, lev):
        res = getattr(s, 'H%s' % lev, s.L)
        return res


class Facts(Cfg):
    """ features config """

    debug = False
    term_width = 80
    no_print = False
    bq_mark = '┃'
    code_mark = ' '
    light_bg = False
    no_smart_indent = False
    horiz_rule = '─'
    single_line_mode = False
    # left and right global indents:
    indent = 0
    rindent = 0
    width = 0  # if set > 0 we set rindent accordingly
    header_numbering = 50  # -1: off, min number of lines to do autonumbering
    header_numb_level_min = 1  # min header level to show the numbers
    header_numb_level_max = 6  # max header level to show the numbers
    header_underlining = '*'  # e.g. '*-' to underline H1 with *** and H2 with ---
    opts_tbl_start = '-'
    opts_tbl_end = ':'

    def __init__(f, md, **kw):
        # first check if the config contains color codes and set to C:
        # now overriding our defaults with kw then with env
        if md.split('\n', 1)[0] == md:
            f.single_line_mode = True
            f.indent = 0
        f.setup(kw)
        f.colr = Colors()
        f.colr.setup(kw)


# ------------------------------------------------ end config - begin rendering
# helper funcs:
def get_subseq_light_table_indent(l0):
    p = '**' if l0.startswith('**') else '*'
    keywrd, l1 = l0[2:].split(p, 1)
    keywrd = l0[:2] + keywrd + p
    l1 = l1
    offs = 1 if l1 and l1[0] == ' ' else 2
    return len(l0) - len(l1[offs:].lstrip()) - (2 * len(p))


def block_quote_status(l, g):
    'blockquote'
    if not l.startswith('>'):
        return 0, l, ''
    _ = l.split(' ', 1)
    lev = len(_[0])
    g['max_bq_depth'] = max(lev, g['max_bq_depth'])
    return lev, _[1], _[0]


h_rules_col = {'-': 'L', '_': 'H3', '*': 'H1'}  # different colors
list_markup = {'- ': ('\x03 ', 'L', '❖ '), '* ': ('\x04 ', 'H2', '▪ ')}
h_rules = '---', '___', '***'


def _main(md, f):
    C, cur_colr = f.colr, 'cur_colr'
    cols = int(f.term_width)
    if f.width:
        f.rindent = cols - f.indent - f.width + f.rindent
    cols = cols - f.indent - f.rindent

    g = {}  # glob parsing state (current color, code blocks)

    # ------------ line tools requiring facts instance, possible ctx g as well:
    def is_opts_tbl(l, b=f.opts_tbl_start, e=f.opts_tbl_end):
        fw = first_word(l)
        if fw and fw.startswith(b) and fw.endswith(e):
            return l.replace(fw, '*%s*' % fw[: -len(e)]), len(fw)
        return l, None

    def is_rule(l):
        if not l[:3] in h_rules:
            return
        ll = len(l)
        return True if l in (ll * '-', ll * '*', ll * '_') else False

    # Line Tools:
    first_word = lambda l: l.split(' ', 1)[0]
    is_header = lambda l: l.startswith('#')
    is_list = lambda l: l.lstrip()[:2] in list_markup
    is_empty = lambda l: l.strip() == ''
    is_md_link = lambda l: l[0] == '[' and 'http' in l and ']' in l

    is_new_block = lambda l: (
        is_header(l)
        or is_list(l)
        or is_opts_tbl(l)[1]
        or is_empty(l)
        or is_md_link(l)
        or l[0] in ('\x02',)
        or is_rule(l)
    )
    # -------------------------------------------------------------------------

    md = md.strip()

    # FENCED CODE BLOCKS:
    # we take them out before all parsing,see http://stackoverflow.com/a/587518
    apo, apos = chr(96), chr(96) * 3  # chr 96 is backtick.
    _ = r'^({apos}[^\n]+)\n((?:[^{apo}]+\n{apo}{apo})+)'.format(apos=apos, apo=apo)
    fncd = re.compile(_, re.MULTILINE)  # finds fenced code
    md = md.replace('\n~~~', apos)  # alternative markup for fenced
    # remembering the blocks by their occurance number (len(g))
    [set(g, len(g), '\n'.join(m.groups()) + apo) for m in fncd.finditer(md)]
    blocks = len(g)
    for i in range(blocks):
        md = md.replace(g[i], '\x02%s' % i)

    g['max_bq_depth'] = 0

    # LINESPROCESSOR:
    lines, out = md.splitlines(), []

    g['header_numbering'] = False
    if f.header_numbering > -1 and len(lines) > f.header_numbering:
        g['header_numbering'] = True
        # init:
        g['header_level'] = dict([(i, 0) for i in range(20)])

    # remove boundary effects:
    lines.insert(0, '')
    lines.append('')

    while lines:

        line = lines.pop(0)
        if is_empty(line):
            out.append('')
            continue
        if debug:
            print(('procesing: ', line))
        if is_rule(line):
            out.append(getattr(C, h_rules_col[line[0]]) + (cols * f.horiz_rule))
            continue

        cb = None  # indentd code blocks:
        while line.startswith('    '):
            cb = cb or []
            cb.append(line[4:])
            line = lines.pop(0)
        if cb:
            if out[-1] == '':
                out.pop()
            g[blocks] = '\n%s\n' % '\n'.join(cb)
            out.append('\x02%s' % blocks)
            blocks += 1
            lines.insert(0, line)
            continue

        ssi = None  # subseq indent for textwrap

        # TEXTBLOCKS: Concat lines which must be wrapped:
        bqm = ''  # blockquote mark. e.g. '>>'.
        bq_lev, line, bqm = block_quote_status(line, g)

        src_line_nr = 0

        # we derive the (static) opts table ssi for a new textblox:
        line, opts_tbl_ssi = is_opts_tbl(line)
        # now we find all other lines belonging to that text block and
        # concat (pop from lines) all of them:
        while lines and not line.endswith('  ') and not is_header(line):

            src_line_nr += 1
            nl, l0 = lines[0], line.lstrip()  # next line, this line

            bqnl = block_quote_status(nl, g)
            if bqnl[0] == bq_lev:
                lines[0] = nl = bqnl[1]  # remove redundant '>'

            elif bqnl[0] != bq_lev and bqnl[0] > 0:
                break  # next line different blockquote level -> new text block

            # finding subseq. indent for textwrap.fill:

            # Little md violation: If first word is starred, we set a ssi to
            # position: first line second word start.
            # Gives easy 2 col wrappable tables when first col is hilited.

            if ssi == None:
                # also violation: we wrap subseq lines with one ':'
                break_lines = False
                if ':' in l0 and ':' in nl:
                    break_lines = True

                if is_list(l0):
                    # replace "- " and "* " with tags:
                    line = list_markup[l0[:2]][0] + l0[2:]
                    ssi = 2
                elif opts_tbl_ssi:
                    ssi = opts_tbl_ssi
                elif l0.startswith('*') and not f.no_smart_indent and src_line_nr == 1:
                    ssi = get_subseq_light_table_indent(l0)

            if is_new_block(nl) or break_lines:
                # line is now one wrapable textblock
                if bqnl[0]:  # block quote new line
                    # adapt next line to parse:
                    lines[0] = (bqnl[2] + ' ') + lines[0]
                break
            else:
                line = line.rstrip() + ' ' + lines.pop(0).lstrip()

        ssi = 0 if ssi is None else ssi
        # lines are now blocks

        g[cur_colr] = C.O  # reset color
        ind = len(line) - len(line.lstrip())
        if bqm:
            bqm += ' '
        line = bqm + line

        if is_header(line):
            cont = line.lstrip('#')
            level = len(line) - len(cont)
            line = cont.lstrip()

            u = getattr(f, 'header_underlining', '')
            if len(u) >= level:
                lines.insert(0, 3 * u[level - 1])

            if g['header_numbering']:
                hl = g['header_level']
                hl[level] = hl.get(level, 0) + 1
                [set(hl, i, 0) for i in hl if i > level]
                nr = '.'.join([str(hl[ll]) for ll in range(1, level + 1)])
                if f.header_numb_level_max > level - 1:
                    if f.header_numb_level_min > 1:
                        nr = nr.split('.')[f.header_numb_level_min - 1 :]
                        nr = '.'.join(nr)
                    if nr:
                        line = nr + ' ' + line

            g[cur_colr] = C.H(level)

        # WRAP:
        if len(line) > cols:
            s = bqm + ' ' * (ind + ssi)
            line = fill(line, subsequent_indent=s, width=cols)
        if is_md_link(line):
            g[cur_colr] = C.GRAY
        out.append(g[cur_colr] + line)

    # --------------- Leaving line/block scanning, reWork complete document now
    g[cur_colr] = C.O
    out = '\n'.join(out)

    # INLINE MARKUP, *, **, backticks
    # Alternating replacements, e.g. code, emph. requires a first space char:
    altern = lambda s, c, r: re.sub(
        r'([^{c}]+){c}([^{c}]+){c}?'.format(c=c), r'\1%s\2%s' % (r, g[cur_colr]), ' ' + s,
    )[
        1:
    ]  # removing space again

    # Star must be replaced, else the re would not work :((
    # currently no way to find single stars and not process them..
    out = out.replace('*', '\x01')
    out = altern(out, apo, C.CODE)  # code
    out = altern(out, '\x01\x01', C.emph)  # **
    out = altern(out, '\x01', C.ital)  # *

    # rearrange resets, to be *before* the line breaks, not after...
    out = out.replace('\n' + C.O, C.O + '\n')
    # ... so that we can look for blockquotes:
    for i in range(g['max_bq_depth'], 0, -1):
        # coloring, take header levels. bq_mark is "|":
        m = ''
        for j in range(1, i + 1):
            m += C.H(j) + f.bq_mark
        m += C.O
        out = out.replace('\n' + '>' * i, '\n' + m)

    # Insert back the stored code blocks:
    code_fmt = lambda c: c.replace('\n', '\n%s%s %s' % (C.L, f.code_mark, C.CODE)).rsplit(
        '\n', 1
    )[0]
    for i in range(blocks):
        out = out.replace('\x02%s' % i, '%s%s%s' % (C.CODE, code_fmt(g[i]), C.O))
    out = out.replace(apos + '\n', '')  # before
    out = out.replace(apos, '')  # after

    for k, v in list(list_markup.items()):
        out = out.replace(v[0], getattr(C, v[1]) + v[2] + C.O)

    out = strip_it(out, C.O)
    if not f.single_line_mode:
        out = '\n' + out + '\n'
    li, ri = f.indent * ' ', f.rindent * ' '
    if li or ri:
        out = li + out.replace('\n', '%s\n%s' % (ri, li))
    out += C.O  # reset
    if not f.no_print:
        print(out)
    return out


def strip_it(out, rst):
    'clumsy way to strip at start at end, including color resets'
    sc = {' ': 1, rst: len(rst), '\n': 1}
    while 1:
        m = False
        for k in sc:
            if out.startswith(k):
                out = out[sc[k] :]
                m = True
            if out.endswith(k):
                out = out[: -sc[k]]
                m = True
            if m:
                break
        if not m:
            break
    return out


def main(md, **kw):
    f = Facts(md, **kw)
    # return _main(md, f), f  # we also return to the client the config
    if debug or f.debug:
        return _main(md, f), f  # we also return to the client the config
    try:
        return _main(md, f), f  # we also return to the client the config
    except Exception as ex:
        print(md)  # clear text
        print(('md error: %s %s ' % (f.colr.CODE, ex)))


def render(md, cols=None, **kw):
    if cols is None:
        cols = get_cols()
    kw['term_width'] = cols
    return main(md, **kw)[0]


def get_help(cols, PY2):
    ff = Facts('\n', term_width=cols)
    md, C = __doc__, ff.colr
    for o in ff, C:
        mmd = ()
        for k, d in sorted(o._parms):
            v = getattr(o, k)
            if o == ff:  # need the perceived len here:
                v = C.H2 + (str('%5s' % str(v)) if PY2 else '%5s' % v) + C.O
            mmd += ('%s %s [%s]' % (v, k, d),)
        md = md % ('\n'.join(mmd))
    return md


# allow to adapt $COLUMNS by setting $term_width:
def get_cols():
    try:
        return int(
            (env('term_width') or os.popen('tput cols 2>/dev/null').read().strip() or 80)
        )
    except:
        return 80


def sys_main():
    import sys

    PY2 = sys.version_info[0] == 2
    if PY2:
        import imp

        imp.reload(sys)
        sys.setdefaultencoding('utf-8')
    import os
    from stat import S_ISFIFO

    err = None
    cols = get_cols()
    if S_ISFIFO(os.fstat(0).st_mode):  # pipe mode
        md = sys.stdin.read()
    else:
        if not len(sys.argv) > 1 or '-h' in sys.argv:
            md = get_help(cols, PY2)
        else:
            md = sys.argv[1]
        if os.path.exists(md):
            with open(md) as fd:
                md = fd.read()
    if err:
        print(err)
        print(md)
    else:
        main(md, term_width=cols)


# ==============================================  Script Formatters ===========


def format_bash(dev_help, cols, lines, script, *args):
    """
    Renders help for a bash script nicely, given it follows some conventions.

    These are:

    1. An `md_doc` function is required, returning general docu as markdown,
    containing the string "<auto_command_doc>"

    2. All public functions must be in this format:

        : 'optional doculines before...'
        function myfunc {
            : 'optional inner doculines (params...)'
            <code lines>
        }

    3. In the command parsing part then this: `show_help $sourced $0 $*`
    with that function elsewhere in your tools:

        show_help () {
            local sourced=$1; shift
            local dev_help=false
            test "${2:-}x" == "make_docx" && { md_doc; exit 0; }
            test "${@: -1}" == "-hh" 2>/dev/null && dev_help=true || {
                test "${@: -1}" == "-h" 2>/dev/null || return 0
            }
            local cols=`stty size | cut -d ' ' -f 2`
            mdvl -f $dev_help $cols "$*"; $sourced && return 1 || exit
        }

    """
    dev_help = True if str(dev_help) in ('True', 'true', '1') else False
    single_func_doc = False
    l = lines
    funcs = []
    start = ": '"
    is_func = lambda l: l.startswith('function ')
    is_cmt_end = lambda l: l.rstrip().endswith("'")
    is_cmt_start = lambda l: l.lstrip().startswith(start)

    def clean(s, head_sub):
        s = s.strip()
        s = s[len(start) :] if s.startswith(start) else s
        s = s[:-1] if s.endswith("'") else s
        s = (('\n' + s).replace('\n#', '\n%s#' % head_sub))[1:]
        return s

    def render_func(m, single_func_doc):
        fn = list(m.keys())[0]
        hf, hs = ('# `Function` **%s**', '##') if single_func_doc else ('### %s', '###')
        nr, pre, post, code = list(m.values())[0]
        md = [hf % fn]
        pre and md.append(clean('\n'.join(pre), hs) + '  ')
        post and md.append(clean('\n'.join(post), hs))
        if code:
            code = '\n'.join(code)
            if post or pre:
                md.append('')
            md.append(code)
            md.extend(['---', ''])
        md = '\n'.join(md)
        return md

    fm = {}
    for i in range(len(l)):
        if is_func(l[i]):
            fn = (l[i] + ' ').split(' ', 2)[1]
            funcs.append({fn: [i, [], [], []]})
            fm[fn] = len(funcs) - 1

    if not '-h' in args[0]:
        match = args[0]
        f = []
        for m in funcs:
            if match in list(m.keys())[0]:
                f.append(m)
        if f:
            funcs = f
            single_func_doc = True

    for m in funcs:
        nr, pre, post, code = list(m.values())[0]
        # pre:
        if is_cmt_end(l[nr - 1]):
            i = nr
            while True:
                i = i - 1
                pre.insert(0, l[i])
                if is_cmt_start(l[i]):
                    break
                if i > 1 and l[i - 1].rstrip().endswith('}'):
                    pre = []  # err
                    break

        # post:
        i = nr
        if is_cmt_start(l[nr + 1]):
            while True:
                i += 1
                post.append(l[i][4:])
                if is_cmt_end(l[i]) and not l[i].strip() == start:
                    break
                if l[i + 1].rstrip().endswith('}'):
                    post = []
                    i = nr  # err
                    break
        if dev_help:
            i += 1
            while True:
                if l[i].strip() == '}':
                    break
                code.append(l[i])
                i += 1

    Facts.indent = 0
    if single_func_doc:
        for m in funcs:
            md = render_func(m, single_func_doc)
            main(md, term_width=cols)
            print()
        return

    # now the full doc. convention is to call with make_doc arg:
    Facts.no_print = True
    Facts.header_numbering = 10
    Facts.header_numb_level_min = 2
    Facts.header_numb_level_max = 2

    full = os.popen(script.split(' ')[0] + ' make_doc').read()
    acd = '<auto_command_doc>'
    full = full.replace(acd, '## Commands\n\n' + acd)
    md = main(full, term_width=cols)[0]

    Facts.header_numbering = -1
    rfuncs, sep = '', '\n\n'
    if dev_help:
        sep = ''
    for m in funcs:
        rfuncs = rfuncs + sep + render_func(m, single_func_doc)
    mdf = main(rfuncs, term_width=cols)[0]
    print((md.replace(acd, mdf)))


def format_file(dev_help, cols, fn, *args):
    if not os.path.exists(fn):
        raise Exception('Not found' + fn)
    with open(fn) as fd:
        lines = fd.read().splitlines()
    if 'bash' in lines[0]:
        format_bash(dev_help, cols, lines, fn, *args)
    else:
        raise Exception('Not supported format')


def run():
    """entry point"""
    import sys

    if len(sys.argv) > 1 and sys.argv[1] == '-f':
        format_file(*sys.argv[2:])
    else:
        sys_main()


if __name__ == '__main__':
    run()



#!/usr/bin/python
# -*- coding: ISO-8859-1 -*-

"""
ansi2html

by BlackJack @ python-forum.de

http://python.sandtner.org/viewtopic.php?p=16273#16273
"""


import re

ansi_re = re.compile(
    '\033\\[' '([\\d;]*)' '([a-zA-Z])'  # Escape characters.  # Parameters.
)  # Command.
#
# Map ANSI colors to HTML 3.2 color names.
#
ansi_color = ('black', 'red', 'green', 'yellow', 'blue', 'purple', 'teal', 'white')


class Ansi2HtmlWriter:
    def __init__(self, normalColor=('black', 'white')):
        self.ret = ''
        self.foreground, self.background = normalColor
        self.open_span = False
        self.ret += '<pre style="color:%s;background:%s">' % normalColor

    def _close_span_tag(self):
        if self.open_span:
            self.ret += '</span>'
            self.open_span = False

    def _write_span_tag(self, style):
        self._close_span_tag()
        if style:
            style_str = ';'.join(
                ['%s:%s' % (name, value) for name, value in style.iteritems()]
            )
            self.ret += '<span style="%s">' % style_str
            self.open_span = True

    def _write(self, data):
        for character, html_escape in (('&', '&amp;'), ('<', '&lt;'), ('>', '&gt;')):
            data = data.replace(character, html_escape)
        self.ret += data

    def writeline(self, line):
        last_end = 0
        for match in ansi_re.finditer(line):
            #
            # Write string before match.
            #
            self._write(line[last_end : match.start()])
            last_end = match.end()

            #
            # Process escape sequence.
            #
            parameters, command = match.groups()

            try:
                parameters = map(int, parameters.split(';'))
            except ValueError:
                parameters = [0]

            #
            # *Set Graphics Rendition* is the only command to handle.
            # The 'blink' effect is not supported.
            #
            if command in 'mM':
                attributes = dict()
                for param in parameters:
                    if param == 0:
                        attributes.clear()
                    elif param == 1:
                        attributes['font-weight'] = 'bold'
                    elif param == 4:
                        attributes['text-decoration'] = 'underline'
                    elif param == 7:
                        attributes['color'] = self.background
                        attributes['background'] = self.foreground
                        self.background, self.foreground = (
                            self.foreground,
                            self.background,
                        )
                    elif 30 <= param <= 37:
                        color = ansi_color[param - 30]
                        attributes['color'] = color
                        self.foreground = color
                    elif 40 <= param <= 47:
                        color = ansi_color[param - 40]
                        attributes['background'] = color
                        self.background = color
                self._write_span_tag(attributes)
        #
        # Write string after last match.
        #
        self._write(line[last_end:])
        self._close_span_tag()
        self.ret += '</pre>'
        return self.ret

    write = writeline

    def writelines(self, lines):
        for line in lines:
            self.write(line)


if __name__ == '__main__':
    testString = '''\033[00m\033[01;34mbin\033[00m
\033[01;34mboot\033[00m
\033[01;36mcdrom\033[00m
\033[01;34mdaten\033[00m
\033[01;34mdev\033[00m
\033[01;34metc\033[00m
\033[01;34mhome\033[00m
\033[01;34minitrd\033[00m
\033[01;36minitrd.img\033[00m
\033[01;36minitrd.img.old\033[00m
\033[01;34mlib\033[00m
\033[01;34mlost+found\033[00m
\033[01;34mmedia\033[00m
\033[01;34mmnt\033[00m
\033[01;34mopt\033[00m
\033[01;34mproc\033[00m
\033[01;34mroot\033[00m
\033[01;34msbin\033[00m
\033[01;34msrv\033[00m
\033[01;34msys\033[00m
\033[01;34mtmp\033[00m
\033[01;34musr\033[00m
\033[01;34mvar\033[00m
\033[01;36mvmlinuz\033[00m
\033[01;36mvmlinuz.old\033[00m'''

    import sys

    try:
        testString = open(sys.argv[1]).read()
    except:
        print('filename as first argument needed')
        sys.exit(0)
    s = Ansi2HtmlWriter()
    print(s.write(testString))



# -*- coding: utf-8 -*-
""" simple conversion functions """
import html
import json
import string
import sys
from collections import OrderedDict
from os import linesep as LS

from theming.tablepretty import PrettyTable
from theming.unicode_chars import UCs

# from ax.utils.six import string_types
string_types = (str,)
ASCIDIGITS = string.ascii_letters + string.digits


def strfilter(s, f):
    # filter s for chars in str f:
    return [x for x in s if x in f]


def printables(s):
    return strfilter(s, string.printable)


def repl_unprintable(res):
    res = res.replace('//', '|_AX//_|')
    for r in (
        ('/r', '\r'),
        ('/n', '\n'),
        ('/t', '\t'),
        ('/COL', ':'),
        ('/COM', ','),
        ('/SPC', ' '),
    ):
        res = res.replace(r[0], r[1])
    res = res.replace('|_AX//_|', '\\')
    return res


def show_unprintable(res):
    """ show line ends and tabs:  \r \n \t \r\n """
    res = res.replace('\r\n', '\\r\\n|_AX_RN_|')
    res = res.replace('\t', '\\t\t')
    res = res.replace('\r', '\\r\r')
    res = res.replace('\n', '\\n\n')
    res = res.replace('|_AX_RN_|', '\r\n')
    return res


def printout(m, get=0):
    if isinstance(m, dict) or isinstance(m, (list, tuple)):
        m = dict_to_txt(m, fmt={'ax': 1})
    if hasattr(sys, 'AXFORMATTER'):
        r = sys.AXFORMATTER.fmt(m)
    else:
        r = m
    if get:
        return r
    print(r)


def mtrim(s, l):
    """ foobazbar -> 'fo..ar' if l=6 """
    s = str(s)
    if len(s) <= l - 2:
        # no need to trim:
        return s
    s = s[: l / 2] + '..' + s[-l / 2 :]
    return s


class PTable(PrettyTable):
    """ reformatting a prettytable with better borders.
    note: without calling print_fancy this class makes no sense """

    encoding = 'UTF-8'

    def __init__(self, *a, **kw):
        """
        set never occurring border chars, then replace back in print_fancy:
        directly replacing the default (|-+) is too risky regarding unwanted
        content replacements
        """
        kw['vertical_char'] = chr(4)
        kw['horizontal_char'] = chr(5)
        kw['junction_char'] = chr(6)
        super(PTable, self).__init__(*a, **kw)
        if not 'float_format' in kw:
            self.float_format = '6.2'
        self.int_format['Nr'] = '3'

    def print_fancy(self, get=0, fancy=1, title=None, dt=None):
        """ replace asci borders with unicode """
        r = str(self)
        v, h, j = '|', '-', '+'
        if fancy:
            v, h, j = UCs.VU, UCs.H, UCs.JUNC
        # insert the final borders, plain or unicode fancies, where also the
        # formatter can color on:
        r = r.replace(chr(4), v).replace(chr(5), h).replace(chr(6), j)
        if title:
            r = '\n\n%s\n\n' % fmt(title.upper(), 'important') + r
        post = ''
        if dt:
            post = fmt('\n%.2fs total runtime' % dt, 'debug')
        if get:
            return fmt(r) + post
        printout(r)
        # no printout (=fmt) for post, since already formatted, the reset codes
        # of formattings of contained chars would kill the main formatting:
        print(post)


def fmt(s, prio='info'):
    if hasattr(sys, 'AXFORMATTER'):
        # fmt(s, 'alarm') -> .fmt_alarm(s) on formatter:
        s = getattr(sys.AXFORMATTER, 'fmt_%s' % prio)(s)
    return s


def dict_to_txt(dictObj, hir=0, fmt={}):
    """
    formats hierarchical lists and maps

    fmt map may contain keys for:
    block_sep<hir>: linebreak between keys at this hir level
    em_<hir>: underline the key at this level (em = emphasis)
    plus general format s (html, wiki currently)

    See supported formats below.
    the 'ax' format is reverse compatible with the parse_text
    module in ../parsing/'

    Features of the ax format:
    See the wiki.


    """
    # do trivial cases here, then client does not have to:
    if fmt == 'dict' or 'dict' in fmt:
        return dictObj
    elif fmt == 'json' or 'json' in fmt:
        return json.dumps(dictObj)

    list_marker = '-'
    assert isinstance(fmt, dict), 'Need the format as dictionary'
    prolog = epilog = ''
    if fmt.get('wiki'):
        nl = ' ||\n'
        ind = ' || '
        sep = ' || '
        start = ' || '

    elif fmt.get('html'):
        """ hirarchical html map of maps - see tests """

        def s(pref1, pref2, hir):
            """ flexible parametrization of styles """
            sm = fmt.get('styles')
            if sm == None or isinstance(sm, string_types):
                # must have a styles MAP:
                return ''
            # checking from most to least specific style settings:
            # fmt = {'html': 1, 'styles': {'td': 'report3'}}) would give all tds
            # report3:
            Style = sm.get(
                '%s%s%s' % (pref1, pref2, hir),
                sm.get(
                    '%s%s' % (pref1, pref2),
                    sm.get('%s' % pref1, '%s%s%s' % (pref1, pref2, hir)),
                ),
            )
            return 'class="%s"' % Style

        tind = '   ' * hir
        prolog = (
            '</td><td>' * min(hir, 1)
            + '\n'
            + tind * 2
            + '<table %s>\n' % s('table', '', hir)
        )

        nl = '</td><td colspan="100" %s></td></tr>\n' % s('td', '3', hir)

        epilog = tind * 2 + '</table>\n%s' % (tind * 2)

        if hir:
            epilog += nl
            nl = '</td><td colspan="100" %s></td></tr>\n' % s('td', '3', hir)
        ind = ''
        sep = '</td><td %s>' % s('td', '2', hir)
        start = '  ' + tind * 2
        start += '<tr %s><td %s>' % (s('tr', '', hir), s('td', '1', hir))

    elif fmt.get('ax'):
        nl = '\n'
        ind = '\t'
        sep = ':\t'
        start = ''

        if fmt.get('style') == 'DBG':
            """ this prints out which format markers are inserted """
            start = 'START' + start
            ind = 'IND' + ind
            sep = 'SEP' + sep
            nl = 'NL' + nl

    else:
        nl = '\n'
        ind = '  '
        sep = ':'
        start = ''

    def draw_str(k, hir):
        return start + ind * hir + '%s' % k + nl

    res = prolog
    block_sep = (nl + start) * fmt.get('block_sep_%s' % hir, 0)
    em = fmt.get('em_%s' % hir)

    if isinstance(dictObj, (list, tuple)):
        # is it a plain list or do we need to separate the structures?
        is_plain = 1
        for k in dictObj:
            if isinstance(k, (tuple, list, dict)):
                is_plain = 0
                break
        lcount = 0
        for k in dictObj:
            lcount += 1
            if not is_plain:
                res += dict_to_txt(k, hir, fmt)
                if lcount < len(dictObj) or len(dictObj) == 1:
                    # insert the separator:
                    res += draw_str(list_marker, hir)
            else:
                res += draw_str(k, hir)

    elif isinstance(dictObj, dict):
        it = list(dictObj.keys())
        if not isinstance(dictObj, OrderedDict):
            it.sort()
        for k in it:
            v = dictObj[k]
            if isinstance(v, (tuple, list, dict)):
                res += block_sep
                res += '%s%s%s%s%s' % (start, ind * hir, k, sep, nl)
                if em:
                    res += start + ind * hir + '-' * (len(str(k)) + 1) + nl
                if isinstance(v, dict):
                    res += start + dict_to_txt(v, hir + 1, fmt=fmt) + block_sep
                else:
                    res += dict_to_txt(v, hir + 1, fmt)
            else:
                res += draw_str('%s%s%s' % (k, sep, v), hir)
                if em:
                    res += start + ind * hir + '-' * (len(str(k)) + 1) + nl
            res += block_sep

    elif isinstance(dictObj, (string_types, bool, int, float)):
        res += draw_str(dictObj, hir)

    else:
        # no exception, just draw the repr:
        res += draw_str(dictObj.__repr__(), hir)
        # raise Exception ("type of %s not supported" % dictObj)
    # we need to stay reverse compatible with the synthesiser,
    # so we have to mark lists as such
    # if res is only a string but input was a list:
    if isinstance(dictObj, (tuple, list)):
        if ''.join(res.rsplit(nl)) == res.rsplit(nl, 1)[0]:
            res = res + draw_str(list_marker, hir)

    return res + epilog


def hstr(v):
    return html.escape(str(v)).replace('\n', '<br>\n')


def obj_to_html_list(iObj, hir=0):
    if isinstance(iObj, (list, tuple)):
        return list_to_html_list(iObj)
    elif isinstance(iObj, dict):
        return dict_to_html_list(iObj)
    else:
        return hstr(iObj)


def dict_to_html_list(dictObj, hir=0):
    pre = LS + '  ' * hir
    res = pre + '<ul>'
    for k, v in list(dictObj.items()):
        # keys must be native types, str, ints..:
        if not isinstance(v, (list, tuple, dict)):
            res += pre + '<li>%s: %s</li>' % (hstr(k), hstr(v))
        else:
            res += pre + '<li>%s:</li>' % hstr(k)
            res += pre + obj_to_html_list(v, hir + 1)
    res += pre + '</ul>' + LS
    return res


def list_to_html_list(listObj, hir=0):
    pre = LS + '  ' * hir
    ret = ''
    if isinstance(listObj, (list, tuple)):
        ret += pre + '<ul>'
        for v in listObj:
            v = obj_to_html_list(v, hir + 1)
            ret += pre + '<li>' + v + '</li>'
        ret += pre + '</ul>' + LS
    return ret


if __name__ == '__main__':
    z = {
        '1': 'v1',
        '2': {'21': 'v1', '222': {'deep': 23, 'deepsub': {'deeper': 33}}, '22': 'v2'},
        '3': 'foo',
    }
    ax = dict_to_txt(z, fmt={'html': 1, 'styles': {'td': 'report3'}})
    ax2 = dict_to_txt(z, fmt={'ax': 1})
    print(ax)
    print(('<pre>%s</pre>' % ax2))
    # other tests in unittests folder
    # see also tests in the parsers folder,
    # dict_to_txt is the reverse function of parse there



#!/usr/bin/env python
"""
OCI Filesystem Creation
"""

# Could be done far smaller.
from datetime import datetime
from devapp import gevent_patched
from re import match
import hashlib
import json
from decorator import decorator
import os
import shutil
from copy import deepcopy
from functools import partial
import subprocess
import requests
from devapp.app import FLG, app, run_app, do, system
from devapp import tools
from operator import itemgetter


from devapp.tools import (
    now,
    exists,
    to_list,
    repl_dollar_var_with_env_val,
    project,
    read_file,
    have_tty,
    write_file,
    walk_dir,
)
import json
import time
from tempfile import mkdtemp
import fnmatch
from hashlib import md5


d_cache = os.path.abspath('.cache')
fn_last_state = d_cache + '/' + 'last_state.json'

FN_CONDA = 'Miniconda3-latest-Linux-x86_64.sh'
cleanups = []

hash_cmd = "fd . '%s' --type file --hidden %s -0 | sort -z | xargs -0 sha1sum"
hash_cmd += ' | sha1sum'


def hash_dir(d, depth=''):
    """hashing all which is relevant incl. files dir, but no other subfolders"""
    if depth:
        depth = '-d %s' % depth
    cmd = hash_cmd % (d, depth)
    h0 = os.popen(cmd).read().split()[0]
    h0 += ''.join([hash_dir(d + '/' + i) for i in os.listdir(d) if 'files' in i])
    return md5(h0.encode('utf-8')).hexdigest()


def make_cache_dir():
    if not exists(d_cache):
        os.makedirs(d_cache)


class Flags:
    autoshort = ''

    class Actions:
        class list:
            d = True
            n = 'Display parsed tree'

            class verbose:
                d = False

        class clear_cache:
            short = 'cc'

        class build:
            class remove:
                n = 'Remove existing images with same name (only before commit of successfully built new one)'
                d = False

        class enter:
            n = 'Enter currently building image with terminal. Disables "run". To run another command than the shell use --cmd. When the cmd is not present we fallback to /bin/sh'
            d = False

            class cmd:
                n = 'Enter currently building image with given command'
                d = '/bin/bash'

        class delete:
            a = True
            n = 'Delete image, i.e. allows rebuild. Will also remove old versions and remove all running containers'

    class name:
        n = 'Exactly match an image name, e.g. "arch_conda"'
        d = ''

    class match:
        n = 'Fuzzy match images, e.g. "cond" (using grep -E)'
        d = ''

    class parents:
        n = 'Include parents of selected image(s) at any operation'
        d = False

    class repeat_from:
        n = 'Repeat last build steps - even if they were succesful. 1: repeat all except first, -2: repeat last 2 successful'
        d = ''

    class mount_home_dir:
        n = 'Mount user home directory at each buildah run'
        d = False

    class yes:
        n = 'All questions: y'
        d = False


g = lambda o, k, d=None: getattr(o, k, d)


class Action:
    _post = cleanups

    def _pre():
        make_cache_dir()
        verify_have_tools()
        if not FLG.clear_cache:
            do(S.load_last_build_cache)
        S.filter_matching()
        S.hash_spec_dirs()
        buildah.images()

    def list():
        return S.dict() if not g(FLG, 'list_verbose') else S.verbose_dict()

    def build():
        # build top down (->sorted):
        for spec in sorted(S.work_tree['specs'], key=lambda i: i['dir']):
            st = spec['status']
            if st == st_current:
                return app.info('Local image is up to date', **spec)
            S.d_cur = d = spec['dir']
            do(img.build, spec)

    def enter():
        pass

    def delete():
        pass

    def clear_cache():
        system('rm -f "%s"' % fn_last_state)


# def run():
#     breakpoint()  # FIXME BREAKPOINT
#     make_cache()
#     verify_have_tools()
#     if FLG.clear_cache:
#         os.system('rm -f "%s"' % fn_last_state)
#     do(S.load_last_build_cache)
#     S.set_matching(match=FLG.name)
#     buildah.images()

#     if FLG.list:
#         return S.dict()

#     if FLG.enter:
#         verify_have_one()
#         return do(img.enter, dir=S.d_cur, cmd=FLG.cmd)

#     if FLG.delete:
#         verify_have_one()
#         do(img.delete_all, dir=S.d_cur)

#     if FLG.run:
#         verify_have_one()
#         cleanups.extend([cleanup_write_state, clean_temp_dir])
#         spec = img.spec(S.d_cur)
#         if FLG.repeat_from:
#             i = int(FLG.repeat_from)
#             actions = spec.get('actions', [])
#             spec['actions'] = actions[:i]

#         if '0' in FLG.run:
#             app.warn('Forgetting build container - rebuilding')
#             spec['C'] = None

#         m = {0: img.set_delete_before_commit, 1: img.build, 2: img.commit, 3: img.run}
#         last = True
#         app.info('Running steps: %s' % FLG.run)
#         for i in [0, 1, 2, 3]:
#             if i == 2 and last == False:
#                 app.info('skipping commit, no build')
#                 continue
#             if str(i) in FLG.run:
#                 last = do(m[i], dir=S.d_cur)
#     return S.dict()


# partial(app.die, example='./my_reg.com[:port]::library/base1/base2')


class State:
    # _delete_before_commit = False
    # by_img = None
    d_root = '.'
    # img_by_dir = None
    _d_tmp = None
    # d_cur = None
    # _selected_spec_dirs = None
    work_tree = None
    # buildah images results:
    _commited_local_images = None

    def __init__(s):
        # s.by_img, s.d_matching, s.img_by_dir, s.work_tree = {}, [], {}, {}
        s.work_tree = {}
        s._d_tmp = mkdtemp()

    def load_last_build_cache(s):
        ls = read_file(fn_last_state, dflt='')
        if ls:
            ls = json.loads(ls)
            [setattr(s, k, v) for k, v in ls.items()]

    def hash_spec_dirs(s):
        for m in S.work_tree['specs']:
            d = m['dir']
            m['hash'] = hash_dir(m['dir'], depth=1)[:6]

    def filter_matching(s):
        """Set to build directory, if match is unique"""
        s.d_cur, s._selected_spec_dirs = False, []

        cs = (
            os.popen('fd -t d | grep -v "\.files" | grep -E "%s"' % FLG.match)
            .read()
            .splitlines()
        )
        w = S.work_tree
        w['filters'] = {'match': FLG.match, 'name': FLG.name}
        dirs = set()
        root_dirs = set(os.listdir())  # registry, not image spec on top level
        for c in cs:
            # c like 'artifacts.foo.com::devapps/arch/arch_conda'
            parts = c.split('/')
            s._selected_spec_dirs.append(c)
            n = parts[-1]
            if not FLG.name or n == FLG.name:
                if not c in root_dirs:
                    dirs.add(c)
            if FLG.parents:
                while '/' in c:
                    c = c.rsplit('/', 1)[0]
                    if not c in root_dirs:
                        dirs.add(c)
        # status fixed for local ones in subsequent images run:
        w['specs'] = [{'dir': i, 'status': 'nonlocal'} for i in sorted(dirs)]

    #        # if len(dirs) == 1:
    #        #     app.info('Exact match -> setting image dir', dir=r[0])
    #        #     s.d_cur = r[0]

    #    # def spec(s, dir=None):
    #    #     d = dir or s.d_cur
    #    #     i = s.img_by_dir[d]
    #    #     return s.by_img[i]

    #    def remember(s, fn, args, kw):
    #        cur = s.spec()
    #        cur.setdefault('actions', []).append(ser_act(fn, args, kw))

    def dict(s):
        g, c = getattr, callable
        return {k: g(S, k) for k in dir(S) if not k.startswith('_') and not c(g(S, k))}

    def verbose_dict(s):
        d = s.dict()
        s.add_git_log(d)
        return d

    def add_git_log(s, d):
        cmd = 'git log --oneline %s --name-status "%s"'
        for s in reversed(d['work_tree']['specs']):
            days, d = s.get('age', 'd')[:-1], s['dir']
            if days:
                days = '--since "%s days ago" ' % days
            gl = cmd % (days, d)
            ls = os.popen(gl).read().strip()
            if not ls:
                app.warn('Not git managed')
                return
            ls = ls.splitlines()
            r = []
            while ls:
                l = ls.pop()
                if '\t' in l[:10] and d in l:
                    l = l.replace(d + '/', '')
                    if '/' in l:
                        if not 'files' in l.split('/', 1)[0]:
                            continue
                r.append(l.replace('\t', ' '))
            s['git log since'] = r

    __repr__ = lambda s: json.dumps(s.dict(), indent=2)


S = State()
# ser_act = lambda fn, args, kw: '%s %s %s' % (fn, str(args), str(kw))

##spec_by_dir = lambda dir: S.by_img.get(S.img_by_dir.get(dir))


die = lambda *a, **kw: app.die(*a, **kw)


def repo_and_ns(s):
    m = {}
    lns = s.split('::')
    if not len(lns) == 2 or not '.' in lns[0]:
        die('first part of dir must be registry:namespace')
    m['repository'] = lns[0]
    m['namespace'] = lns[1]
    return m


# @decorator
# def remembered(func, maxms=60000, *args, **kw):
#    """Writing image building steps into our log which is written at exitted
#    So that we can remember what was already run for a certain build and skip it.
#    """
#    remb = kw.pop('remember', True)
#    t0 = now()
#    fn = func.__name__
#    sa = ser_act(fn, args, kw)
#    if remb:
#        for a in S.spec().get('actions', []):
#            if a == sa:
#                return app.warn('skip: %s' % fn, args=args, **kw)
#    result = func(*args, **kw)
#    if not remb:
#        return result
#    S.remember(fn, args, kw)
#    S._dt_last_cmd = dt = now() - t0
#    w = app.warn if dt > maxms else app.info
#    w('%s took %d ms' % (fn, dt))
#    return result


B = 'buildah '

strptime = datetime.strptime
unix0 = datetime(1970, 1, 1)

st_current, st_outdated, st_unknown = 'current', 'outdated', 'unknown'


class buildah:
    def images():
        now = time.time()
        p, j = os.popen, json.loads
        l = S._commited_local_images = j(p(B + 'images --json').read().strip()) or []

        def gb(s):
            s, u = s.split()
            return float(s) if u == 'GB' else float(s) / 1024.0 if u == 'MB' else 0

        m = {'count': len(l), 'total': '%s GB' % int(sum([gb(i['size']) for i in l]))}
        app.info('Local images', **m)
        specs = S.work_tree['specs']
        S.work_tree['local images'] = m
        for i in l:
            ns = i['names']
            ns = [n.split(':') for n in ns if not ':latest' in n or len(ns) < 2]
            ns = sorted(ns, key=itemgetter(1))[-1]
            f = '%Y-%m-%dT%H:%M:%S'
            utc_dt = datetime.strptime(i['createdatraw'].split('.', 1)[0], f)
            ts = i['ts'] = (utc_dt - unix0).total_seconds()
            n = i['name'] = ns[0].rsplit('/', 1)[-1]
            t = i['tag'] = ns[1]
            s = [k for k in specs if k['dir'].endswith('/' + n)]
            assert len(s) < 2
            if s:
                s = s[0]
                s['age'] = '%sd' % int((now - ts) / 86400)
                s['status'] = st_current if s['hash'] in t else st_outdated
                s['size'] = i['size']

    def test_build_avail(C):
        if not C:
            return
        if os.popen(B + 'run %s -- ls /' % C).read().strip():
            app.info('build present', cont=C)
            return C


#    def from_(spec):
#        cmd = 'buildah from "%(from)s"' % spec
#        app.debug(cmd)
#        C = os.popen(cmd).read().strip()
#        app.info('container', container_name=C)
#        return C

#    def pull(img, **kw):
#        # err = do(system, 'buildah pull "%s"' % img, no_fail=True)
#        return do(system, 'buildah pull "%s"' % img, **kw)

#    def copy(frm, to):
#        # updated always
#        cmd = 'buildah copy --add-history "%s" "%s" "%s"' % (S.C, frm, to)
#        app.info(cmd)
#        do(system, cmd)

#    def exists(fn):
#        cmd = B + 'run %s -- ls "%s" 2>/dev/null 1>/dev/null'
#        return not os.system(cmd % (S.C, fn))

#    def has_command(cmd):
#        cmd = B + 'run %s -- /bin/sh -c "command -v %s" 2>/dev/null 1>/dev/null'
#        return not os.system(cmd % (S.C, cmd))

#    @remembered
#    def run(cmd, add_history=True, **kw):
#        app.info('%s (inside)' % cmd.split(' ')[0].replace('/.work/', ''), cmd=cmd)
#        h = '--add-history' if add_history else ''
#        cmd = 'buildah run %s %s %s -- %s' % (volmounts(), h, S.C, cmd)
#        app.info(cmd)
#        do(system, cmd)

#    @remembered
#    def env(vars):
#        e = ' '.join([' -e %s=%s ' % (k, v) for k, v in vars.items()])
#        do(system, 'buildah config --add-history %s %s' % (e, S.C))

#    @remembered
#    def entrypoint(ep):
#        b = 'buildah config'
#        do(system, b + ''' --add-history --entrypoint '[ "%s" ]' %s''' % (ep, S.C))

#    @remembered
#    def add_to_buildenv(vars):
#        fn = S._d_tmp + '/build_env.step'
#        s = '\n'.join(['export %s="%s"' % (k, v) for k, v in vars.items()])
#        write_file(fn, s)
#        buildah.copy(fn, '/.work/build_env.step')
#        add_script('add_to_buildenv.sh', add_to_buildenv)
#        buildah.run('/.work/add_to_buildenv.sh')

#    @remembered
#    def commit(img, build_img):
#        x = datetime.now()
#        tagged = '%s:latest %s:%s.%s.%s' % (img, img, x.year, x.month, x.day)
#        if S._delete_before_commit:
#            for k in img, tagged:
#                do(
#                    system,
#                    'buildah rmi -f "%s" 2>/dev/null; echo ok' % k,
#                    no_fail=True,
#                )
#        do(system, 'buildah commit %s %s' % (build_img, img))
#        do(system, 'buildah tag ' + tagged)


# def volmounts():
#    v = ''
#    if FLG.mount_home_dir:
#        v = ' -v "%(HOME)s:%(HOME)s"' % os.environ
#    return v


# def add_script(name, script=None):
#    dir = '/.work'
#    if name.startswith('/'):
#        dir, name = name.rsplit('/', 1)
#    fn = S._d_tmp + '/%s' % name
#    if script:
#        write_file(fn, script, chmod=0o755)
#    else:
#        os.system('chmod 755 "%s"' % name)
#    buildah.copy(fn, dir + '/' + name)


# class steps:
#    """Steps we log in the cache and won't repeat if done already """

#    def copy_workdir(dir, **kw):
#        """
#        Copying an image definition dir (w/o it't sub img definitions) into a container.

#        - We go via the tempdir on the host, so that we can delete the sub img folders.
#        - We detect runnable scripts and chmod +x them, plus add #!/bin/sh if not given
#        """
#        cmd = 'cp -a "%s/"* "%s/"' % (dir, S._d_tmp)
#        do(system, cmd)
#        for sub in os.listdir(S._d_tmp):
#            fn = S._d_tmp + '/' + sub
#            if os.path.isdir(fn) and sub[0] == ':':
#                os.system('rm -rf "%s"' % fn)
#            if is_runnable(sub):
#                s = read_file(fn)
#                if not s.startswith('#!'):
#                    app.info('Adding #!/bin/sh', fn=sub)
#                    s = '#!/bin/sh\n' + s
#                l = s.split('\n', 1)
#                s = '%s\n\n. /.work/build_env\n. /enter_env\n%s' % (l[0], l[1])
#                write_file(fn, s, chmod=0o755)
#        buildah.copy(S._d_tmp, '/.work')
#        buildah.run('touch /.work/build_env')
#        app.debug('copied workdir into image')


# is_runnable = lambda n: n.endswith('.sh') or '.run:' in n


class img:
    def build(spec):
        """Recursively pull or build parents"""
        img.complete_spec(spec)

        for step in sorted(os.listdir(spec['dir'])):
            breakpoint()  # FIXME BREAKPOINT
            if not ':' in step:
                step += ':'
            typ, n = step.split(':', 1)
            if not '.' in typ:
                continue
            nr, mode = typ.split('.', 1)
            if not nr.isdigit():
                continue
            mode = {'from': 'from_'}.get(mode, mode)
            # if not spec['local']['build'] and not mode == 'from_':
            # from is the first - if present. If not, we use the parent
            if not mode == 'from_':
                do(img.pull_or_build_parent, spec=spec)
                modes.from_('', spec)
            moderunner = getattr(modes, mode, None)
            if not moderunner:
                app.die('Mode not supported', step=step, mode=mode)
            r, p = repl_dollar_var_with_env_val(step, ask_on_fail=True, get_vals=True)
            if p:
                buildah.add_to_buildenv(p)
            app.info('build step: %s' % step)
            breakpoint()  # FIXME BREAKPOINT
            moderunner(step, spec)

    def complete_spec(spec):
        """Analyse and take appart a directory in the tree, adding infos which list did not show"""
        breakpoint()  # FIXME BREAKPOINT
        m = spec
        # if m:
        #     S.C = m.get('C')
        #     img.check_is_local(m)
        #     return m
        dir = m['dir']
        app.debug('analyzing', dir=dir)
        l = dir.split('/')
        m['name'] = l[-1]
        m.update(repo_and_ns(l[0]))
        frm = [k for k in os.listdir(dir) if '.from:' in k]
        m['img'] = '%(repository)s/%(namespace)s/%(name)s' % m
        if len(frm) > 1:
            app.die('More than 1 from statements', dir=dir, frms=frm)
        if not frm:
            m['dir_parent'] = dp = '/'.join(l[:-1])
            p = img.spec(dp)
            m['from'] = p['img']
        else:
            s = frm[0].split('.from:', 1)[1].rsplit(':', 1)
            m['from'] = s[0].replace('::', '/') + '/' + s[1]
        i = m['img']
        m['version'] = 'latest'
        m['from'] += ':%(version)s' % m
        img.check_is_local(m)
        app.debug(i, **m)
        m['ts_analyzed'] = now()
        S.by_img[i] = m
        S.img_by_dir[dir] = i
        S.C = m.get('C')
        return m

    #    # def spec(dir):
    #    #     """Analyse and take appart a directory in the tree"""
    #    #     m = spec_by_dir(dir)
    #    #     if m:
    #    #         S.C = m.get('C')
    #    #         img.check_is_local(m)
    #    #         return m
    #    #     app.debug('analyzing', dir=dir)
    #    #     l = dir.split('/')
    #    #     m = {'dir': dir, 'name': l[-1]}
    #    #     m.update(repo_and_ns(l[0]))
    #    #     frm = [k for k in os.listdir(dir) if '.from:' in k]
    #    #     m['img'] = '%(repository)s/%(namespace)s/%(name)s' % m
    #    #     if len(frm) > 1:
    #    #         app.die('More than 1 from statements', dir=dir, frms=frm)
    #    #     if not frm:
    #    #         m['dir_parent'] = dp = '/'.join(l[:-1])
    #    #         p = img.spec(dp)
    #    #         m['from'] = p['img']
    #    #     else:
    #    #         s = frm[0].split('.from:', 1)[1].rsplit(':', 1)
    #    #         m['from'] = s[0].replace('::', '/') + '/' + s[1]
    #    #     i = m['img']
    #    #     m['version'] = 'latest'
    #    #     m['from'] += ':%(version)s' % m
    #    #     img.check_is_local(m)
    #    #     app.debug(i, **m)
    #    #     m['ts_analyzed'] = now()
    #    #     S.by_img[i] = m
    #    #     S.img_by_dir[dir] = i
    #    #     S.C = m.get('C')
    #    #     return m

    def check_is_local(spec):
        breakpoint()  # FIXME BREAKPOINT
        m = spec.setdefault('local', {})
        m['build'] = b = bool(buildah.test_build_avail(spec.get('C')))
        if b:
            app.info('%s' % spec['C'])
        m['img'] = f = img.is_local(spec['img'])
        m['from'] = i = img.is_local(spec['from'])
        app.info('local', **m)
        # if not b:
        # app.debug('resetting build actions')
        # spec['actions'] = []

    #    # def pull_or_build_parent(spec):
    #    #     if img.is_local(spec['from']):
    #    #         return
    #    #     err = do(buildah.pull, img=spec['from'], no_fail=True)
    #    #     if err:
    #    #         do(img.build, dir=spec['dir_parent'])
    #    #     buildah.images()
    #    #     assert img.is_local(spec['from'])

    def is_local(n):
        breakpoint()  # FIXME BREAKPOINT
        for h in S._commited_local_images:
            ns = h.get('names') or ()
            if n in ns or n + ':latest' in ns:
                return True
        return False


#    # def enter(dir, cmd):
#    #     m = S.spec(dir=dir)
#    #     app.info('Entering container', json=m)
#    #     app.warn('Your changes will NOT be recorded!')
#    #     if not buildah.has_command(cmd):
#    #         app.warn('Missing inside', cmd=cmd, fallback='/bin/sh')
#    #         cmd = '/bin/sh'
#    #         app.warn('No bash - using sh')
#    #     buildah.run(cmd, add_history=False, remember=False)
#    #     app.info('exitted', cont=m['C'])

#    # def commit(dir):
#    #     spec = img.spec(dir)
#    #     if not spec['local']['build']:
#    #         return app.info('nothing to commit', reason='no build')
#    #     do(buildah.commit, img=spec['img'], build_img=S.C)
#    #     system('buildah images | grep "%s"' % spec['img'])

#    # def set_delete_before_commit(dir):
#    #     spec = img.spec(dir)
#    #     S._delete_before_commit = True

#    # def run(dir):
#    #     spec = img.spec(dir)
#    #     if not spec['local']['img']:
#    #         app.die('Not present', img=img)
#    #     c = 'podman' if not os.system('command -v podman') else 'docker'
#    #     do(system, c + ' run -ti --rm %s "%s" /bin/bash' % (volmounts(), spec['img']))

#    # def delete_all(dir):
#    #     spec = img.spec(dir)
#    #     r = []
#    #     for i in S._commited_local_images:
#    #         if any([spec['img'] in n for n in (i.get('names') or ())]):
#    #             r.append(i)
#    #     if not r:
#    #         return
#    #     app.info('Will remove', json=r)
#    #     if have_tty() and not FLG.yes:
#    #         if not 'y' in input('Ok to delete [y/N]? ').lower():
#    #             app.die('unconfirmed')
#    #     for i in r:
#    #         do(system, 'buildah rmi -f "%s"' % i['id'])


# class modes:
#    def from_(step, spec):
#        S.C = spec['C'] = buildah.from_(spec=spec)
#        do(steps.copy_workdir, **spec)
#        if not buildah.exists('/enter_env'):
#            add_script('/enter_env', enter_env)
#            add_script('/enter.sh', enter)
#            buildah.env({'LC_ALL': 'en_US.UTF-8', 'LANG': 'en_US.UTF-8'})
#            buildah.entrypoint(('/enter.sh'))
#        spec['local']['build'] = True

#    def pkgs(step, spec, typ='distri'):
#        u = 'updated' if 'updated:' in step else ''
#        step = step.replace('updated:', '')
#        args = '"' + (step.split(':', 1)[1].replace('+', '" "') + '"')
#        add_script('pkg_inst.sh', pkg_inst)
#        buildah.run(cmd='/.work/pkg_inst.sh %s %s %s' % (typ, u, args))

#    def pip(step, spec):
#        return modes.pkgs(step, spec, typ='pip')

#    def yarn(step, spec):
#        return modes.pkgs(step, spec, typ='yarn')

#    def conda(step, spec):
#        return modes.pkgs(step, spec, typ='conda')

#    def add_user(step, spec):
#        add_script('useradd.sh', useradd)
#        username = step.split(':')[1]
#        buildah.run(cmd='/.work/useradd.sh %s' % username)

#    def run(step, spec):
#        buildah.run(cmd='/.work/%s' % step)

#    def files(step, spec):
#        if step.endswith(':'):
#            step = step[:-1]
#        s = ['#!/bin/sh', '( cd "/.work/%s/" && tar cf - . ) | tar xfv - -C /' % step]
#        add_script(step + '.sh', '\n'.join(s))
#        buildah.run(cmd='/.work/%s.sh' % step)

#    def add_path(step, spec):
#        P = step.split(':', 1)[-1]
#        s = ['#!/bin/sh']
#        for p in P.split('+'):
#            p = '/' + p.replace('__', '/')
#            s.append('''echo -e 'export PATH="%s:$PATH"' >> /enter_env''' % p)
#        add_script('add_path.sh', '\n'.join(s))
#        buildah.run('/.work/add_path.sh')

#    def install_conda(step, spec):
#        fn = FN_CONDA
#        ffn = d_cache + '/' + fn
#        if not buildah.exists('/.work/conda.sh'):
#            if not exists(ffn):
#                cmd = 'curl "https://repo.anaconda.com/miniconda/%s" > "%s"' % (fn, ffn)
#                if os.system(cmd):
#                    app.die('could not download conda')
#            os.system('chmod +x "%s"' % ffn)
#            buildah.copy(ffn, '/.work/conda.sh')
#        add_script('conda_inst.sh', conda_inst)
#        buildah.run(cmd='/.work/conda_inst.sh')

#    # def is_runnable(step, spec):
#    #     cmd = '/.work/' + step
#    #     app.info(step)
#    #     system('cat "%s/%s"' % (spec['dir'], step))
#    #     buildah.run(cmd=cmd)


# def cleanup_write_state():
#    print('dumping state', fn_last_state)
#    write_file(fn_last_state, json.dumps(S.dict(), indent=4), mkdir=True)


# def clean_temp_dir():
#    d = S._d_tmp
#    if exists(d) and ('tmp' in d or 'temp' in d):
#        os.system('rm -rf "%s"' % d)


def verify_have_tools(required=['buildah', 'fd', 'curl']):
    miss = lambda t: os.system('command -v %s >/dev/null' % t)
    [app.die('Required tool missing', tool=t) for t in required if miss(t)]


# def verify_have_one():
#    if not S.d_cur:
#        m = 'Require exactly one matching image (d_cur)'
#        app.die(m, found=len(S._selected_spec_dirs))


main = lambda: run_app(Action, flags=Flags)


if __name__ == '__main__':
    main()

# enter_env = '''
# export LC_ALL=en_US.UTF-8
# export LC_CTYPE=en_US.UTF-8
# export LANG=en_US.UTF-8
# export LANGUAGE=en_US.UTF-8
#'''


# enter = '''#!/bin/sh
# . /enter_env
# eval "$@"
#'''
# add_to_buildenv = '''#!/bin/sh
# cat /.work/build_env.step >> /.work/build_env || true
#'''

# useradd = '''#!/bin/sh
# set -e
# . /.work/build_env || true
# username="$1"
# userdel $username && rm -rf /home/$username || true
# useradd -m -u 1000 -g users -d /home/$username -s /usr/bin/fish $username
# command -v sudo && echo "$username ALL=(ALL) ALL" >> /etc/sudoers
# echo -e ''$userpw'\n'$userpw'' | passwd $username
#'''

# pkg_inst = '''#!/bin/sh
# set -ex
# . /.work/build_env || true

# function set_packager {
#    command -v "apt-get" && { pkger="apt"; return; }
#    command -v "yum"     && { pkger="yum"; return; }
#    command -v "pacman"  && { pkger="pacman"; return; }
#    echo "no package manager"
#    exit 1
# }

# function update_pacman {
#   echo "updating pacman"
#   pacman --noconfirm -Syu
# }

# function install_pacman {
#   pacman --noconfirm -S "$@"
# }


# function update_yum {
#   echo "updating yum"
#   yum -y update
# }

# function install_yum {
#   yum -y install "$@"
# }

# function update_apt {
#   echo "updating apt"
#   apt-get -y update
# }

# function install_apt {
#   apt-get -y install "$@"
# }

# function install_yarn {
#    yarn global add "$@"
# }

# function install_pip {
#    pip install "$@"
# }

# function install_conda {
#    conda install -y "$@"
# }

# function main {
#    . /enter_env # possibly path -> e.g. pip elsewhere
#    if [ "$1" == "distri" ]; then
#        set_packager
#    else
#        pkger="$1"
#    fi
#    shift
#    test "$1" == updated && { update_$pkger; shift; }
#    install_$pkger "$@"

# }


# main "$@"
#'''

# conda_inst = '''#!/bin/sh
# set -x
# rm -rf /opt/conda
# /.work/conda.sh -b -p /opt/conda
# rm -f /etc/profile.d/conda.sh
# ln -s /opt/conda/etc/profile.d/conda.sh /etc/profile.d/conda.sh
# echo ". /opt/conda/etc/profile.d/conda.sh" >> ~/.bashrc
# echo "conda activate base" >> ~/.bashrc
# find /opt/conda/ -follow -type f -name '*.a' -delete
# find /opt/conda/ -follow -type f -name '*.js.map' -delete
# /opt/conda/bin/conda clean -afy
# echo -e '
# export PATH="/opt/conda/bin:$PATH"
# export CONDA_EXE="/opt/conda/bin/conda"
# export CONDA_PREFIX="/opt/conda"
# export CONDA_PROMPT_MODIFIER="(base)"
# export CONDA_SHLVL="1"
# export CONDA_PYTHON_EXE="/opt/conda/bin/python"
# export CONDA_DEFAULT_ENV="base"
# export _CE_M=
# export _CE_CONDA=
# export _CONDA_SHLVL="1"
#' >> /enter_env
#'''



#!/usr/bin/env python
"""
File is called by the shell when a test script is to be run.
Our job: Embed it into wanted environ.
sys.argv[1]: the test script
Rest in environ
"""

import os
import pdb
import sys
import time
from functools import partial

breakpoint = pdb.set_trace

env = os.environ

# print('ing build env', os.environ)
# print('args', sys.argv)
# print('env', os.environ)

d_repo = os.path.abspath(os.curdir)
d_rwdir = d_repo + '/.devapp_rw_top_layer'
d_assets = '/assets'
d_rootfs = d_assets + '/rootfs'
d_rootfss = os.listdir(d_rootfs)
dt = lambda: now() - t0
now = lambda: int(time.time() * 1000)
uc = lambda s: s.upper()
nosl = lambda s: s.replace('/', '-')
listed = (
    lambda s: list(s)
    if isinstance(s, (tuple, list))
    else [p.strip() for p in s.split(',')]
)
pretty_d = lambda d: ', '.join(['%s: %s' % (k, str(v)) for k, v in d.items()])
commit_msg = lambda: env['CI_COMMIT_MESSAGE']
d_abs_checkout = lambda: env['CI_PROJECT_DIR']
t0 = now()


class Log:
    def msg(self, msg, **kw):
        print('%6s %6s %s %s' % (dt(), uc(kw.pop('level')), msg, pretty_d(kw)))


log = Log()
[setattr(log, l, partial(log.msg, level=l)) for l in ['debug', 'info', 'warn', 'error']]
# replace this with a check for a spec:
def fs_stack():
    s = env.get('dflt_fs_stack')
    if not s:
        log.warn('No dflt_fs_stack defined in this runner - picking "centos7"')
    return s or 'centos7'


mount_tmpl = 'mount -t overlay -o lowerdir="%s",upperdir="%s",workdir="%s" '
mount_tmpl += 'overlay "%s"'


def _union(lower, upper, dir_name, d_build):
    # http://blog.programster.org/overlayfs
    D = d_stack() + '/%s' % dir_name
    os.mkdir(D)
    os.mkdir(D + '.work')
    cmd = mount_tmpl % (lower, upper, D + '.work', D)
    runproc(cmd)
    ctx['mounts'].append(D)
    return D


def abs_dir(d):
    if d.startswith('/'):
        raise Exception('Sorry no absolute dirs')
    if d in d_rootfss:
        return d_rootfs + '/' + d
    raise Exception('Not available %s' % d)


# this keeps the stack dir - in order to cleanup even at crashes:
ctx = {'mounts': []}


def main(script, *args):
    if not '.git' in os.listdir(d_repo):
        raise Exception('Not in a checkout dir')
    if not os.path.exists(d_repo + '/' + script):
        raise Exception('Not found: %s' % script)

    stack = listed(fs_stack())

    # contains all done with the fsstack during test:
    if not os.path.exists(d_rwdir):
        os.mkdir(d_rwdir)
    log.debug('Building fs stack', stack=stack)

    stack.append(d_rwdir)
    env['dev_apps_top_layer'] = d_rwdir

    # must be within the repo, it is itself mounted by gitlab:
    ctx['d_stack'] = d_abs_checkout() + '/.devapp_fs_stack'

    os.mkdir(d_stack())

    make_union = partial(_union, d_build=d_stack())
    cur = abs_dir(stack.pop(0))
    dir_name = 0
    while stack:
        n = stack.pop(0)
        dir_name += 1
        dir_name, n = ('final', n) if not stack else (dir_name, abs_dir(n))
        cur = union(cur, n, dir_name=dir_name)

    # lastely we mount the repo into it:
    dn = d_abs_checkout()
    dnf = '%s/final%s' % (ctx['d_stack'], dn)
    log.debug('Mounting repo in', into=dnf)
    bindmount(d_abs_checkout(), dnf)

    make_script(script, *args)
    run_script()


def make_script(script, *args):
    a = ['#!/usr/bin/env bash']
    a.append('set -a')
    for k, v in os.environ.items():
        a.append("%s='%s'" % (k, v))
    a.append('set +a')
    a.append('cd "$CI_PROJECT_DIR"')
    a.append('echo -e "\\033[38;5;82;1m * Running %s: \\033[0;m \\n"' % script)
    a.append('%s %s' % (script, ' '.join(args)))
    a = '\n'.join(a)
    ts = '%(d_stack)s/final/testscript' % ctx
    with open(ts, 'w') as fd:
        fd.write(a)
    runproc('chmod +x "%s"' % ts)
    log.debug('Entry script written')  # , script=a)


def run_script():
    D = ctx['d_stack'] + '/final'
    log.info('Running test container')
    runproc('systemd-nspawn -M $CI_COMMIT_SHA -D "%s" /testscript' % D)


def bindmount(frm, to):
    runproc('mkdir -p "%s"' % to)
    runproc('mount -o bind "%s" "%s"' % (frm, to))
    ctx['mounts'].append(to)
    # print(str(os.listdir(d_stack() + '/final')))
    # print(cur)


def runproc(cmd):
    cmd = cmd.replace(d_abs_checkout(), '$CI_PROJECT_DIR')
    log.debug('Running', cmd=cmd)
    r = os.system(cmd)
    if r:
        raise Exception('Cmd failed: %s' % cmd)


d_stack = lambda: ctx.get('d_stack')


def cleanup():
    """umount all overlays then remove the stackdir"""
    log.info('Cleanup', mounts=ctx['mounts'])
    while ctx['mounts']:
        m = ctx['mounts'].pop()
        runproc('umount "%s" 2>/dev/null' % m)
    runproc('/bin/rm -rf "%s"' % d_stack())


if __name__ == '__main__':
    try:
        main(sys.argv[1], *sys.argv[2:])
    except Exception as ex:
        raise ex
    finally:
        cleanup()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ax',
 'ax.utils',
 'ax.utils.ax_tree',
 'axlib',
 'axlib.ax_tree',
 'devapp',
 'devapp.app_token',
 'devapp.arch',
 'devapp.arch.gitlab',
 'devapp.arch.gitlab.solution.conf',
 'devapp.components',
 'devapp.lib',
 'devapp.operations',
 'devapp.plugins.ops_devapp',
 'devapp.plugins.ops_devapp.arch',
 'devapp.spec',
 'devapp.testing',
 'devapp.tests',
 'devapp.utils',
 'mdvl',
 'structlogging',
 'structlogging.tests',
 'theming',
 'theming.filesize',
 'theming.formatting',
 'theming.tests',
 'tree_builder',
 'tree_builder.arch']

package_data = \
{'': ['*'],
 'devapp': ['third/*'],
 'devapp.arch.gitlab': ['solution/*'],
 'devapp.spec': ['templates/*']}

install_requires = \
['PyJWT',
 'PyYAML',
 'Pygments',
 'Rx',
 'absl-py',
 'bpytop',
 'decorator',
 'gevent',
 'httpx',
 'humanize',
 'inflection',
 'jsondiff',
 'lz4',
 'msgpack',
 'psutil',
 'pycond',
 'requests',
 'snoop',
 'structlog',
 'tabulate',
 'toml',
 'ujson']

entry_points = \
{'console_scripts': ['app = devapp.plugin_tools:main',
                     'ops = devapp.plugin_tools:main']}

setup_kwargs = {
    'name': 'devapps',
    'version': '2022.1.18',
    'description': 'Apps - End to End.',
    'long_description': '\n# About `devapp`\n\n\n<p attr="autogenerated by make_badges"></p>\n\n<p attr="autogenerated by make_badges"></p>\n\n![](img/page-teaser.png){: style="width:1550px" }\n\n\nThis is an enabler repo to develop "ops friendly" apps in a normalized way, incl:\n\n- logging ([structlog](https://www.structlog.org/en/stable/))\n- cli flags handling ([abseil](https://abseil.io/docs/python/), with addons)\n- docu ([mkdocs-material](https://squidfunk.github.io/mkdocs-material/getting-started/))\n- project setup\n- (test) resoures management, including daemons and container filesystem layers\n\nand more.\n\n\n\n<!-- pre_proc_marker -->\n\n\n----\n\n\nLast modified: Wed Aug 11 17:33:48 2021',
    'author': 'Gunther Klessinger',
    'author_email': 'g_kle_ss_ing_er@gmx.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://axgkl.github.io/devapps',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import absolute_import, division, print_function

from .term import I, L, M, R

"""Some Terminal Messages used repeatedly"""


def DeprecationMessage(msg, ticket=None, alternatives=(), eol=None):
    s = ', '.join([M(a) for a in alternatives])
    if s:
        s = 'Alternatives: ' + s
    if eol:
        s = 'End of live: %s' % I(eol)
    if ticket:
        s += ' ' + L('[%s]' % ticket)
    return '\n'.join([R('%s' % msg.upper()), '', s])



from inspect import getsource


class Mkdocs:
    details = '''

<details>
    <summary>%s</summary>

%s

</details>

'''
    code_ = '''

```_code_
%s
```

    '''
    admon_ = '''

!!! %s "%s"

%s 

    '''
    admon_closed_ = admon_.replace('!!!', '???')
    admon_clsabl_ = admon_.replace('!!!', '???+')
    tab_ = '''
=== "%s"

%s

    '''
    js = code_.replace('_code_', 'js')
    py = code_.replace('_code_', 'python')

    code = lambda which, body: Mkdocs.code_.replace('_code_', which) % body
    tab = lambda title, cont, t=tab_: t % (title, indent(cont))
    admon = lambda title, cont, mode='note', widg=admon_: widg % (
        mode,
        title.replace('\n', ''),
        indent(cont),
    )
    closed_admon = lambda *a, _=admon, __=admon_closed_, **kw: _(*a, widg=__, **kw)
    clsabl_admon = lambda *a, _=admon, __=admon_clsabl_, **kw: _(*a, widg=__, **kw)


def to_min_header_level(l, s):
    h = '\n'
    s = '\n' + s
    for k in range(1, l):
        h += '#'
        s = s.replace(h + ' ', ('\n' + '#' * l) + ' ')
    return s


indent = lambda s: ('\n' + s).replace('\n', '\n    ')[1:]


def header(s, level):
    return level * '#' + ' ' + s.strip()


def deindent(
    s, add_code_line_seps=False, dl=lambda l, ws: l[ws:] if l[:ws] == ' ' * ws else l
):
    # can handle first line not indentend as in docstrings starting after """ w/o line sep
    s = s or ''
    if not s.strip():
        return s
    ls = s.splitlines()
    # first line might be special after like '''def foo'''
    # if indented though we use it:
    if len(ls) == 1:
        return ls[0].strip()
    if ls[0].startswith(' ') or ls[0][:4] in ('def ', 'clas'):
        i = 0
    else:
        for i in 1, 2, 3, 4:
            if i > len(ls) - 1:
                i = i - 1
                break
            if ls[i].lstrip():
                break
    # len of indent to remove:
    ws = len(ls[i]) - len(ls[i].lstrip())
    ls = [dl(l, ws) for l in ls]
    if add_code_line_seps:
        ls = line_sep_before_code(ls)
    return '\n'.join(ls).strip()


def line_sep_before_code(ls):
    """for markdown"""
    r, is_code = [], False
    for ln in ls:
        if not is_code:
            if ln.startswith('    '):
                r.append('')
                is_code = True
        else:
            if not ln.startswith('    '):
                is_code = False
        r.append(ln)
    return r


# ----------------------------------------------------- Create markdown from whole tree
def extract_docstr_head(docstr):
    """
    Finding Docstrings

    Example: The one of this func would be 'Finding Docstrings'  
    """
    s, rest = docstr.strip(), ''
    if not s:
        return '', ''
    l = s.splitlines()
    h = len(l[0]) < 80 and (len(l) < 2 or not l[1].strip())
    if len(l[0]) < 80 and (len(l) < 2 or not l[1].strip()):
        rest = '\n'.join(l[1:]).strip()
        l0 = l[0]
        if l0.startswith('#'):
            # remove markdown:
            return l0.split(' ', 1)[1], rest
        return l0, rest
    # no docstring with header:
    return '', docstr


def obj_tree_to_markdown(obj, hir=1):
    """obj as created by auto_docs recurse_into"""
    T = Mkdocs
    r = []
    add = r.append
    if 1:
        n = obj['name']
    head, rest = extract_docstr_head(obj['doc'])
    add(header('`%s` %s' % (n, head), hir))
    add(rest)
    src = obj['source']
    src = T.code('python', src)
    add(Mkdocs.closed_admon('source', src))
    cs, fs = obj.get('classes', ()), obj.get('funcs', ())
    if cs:
        add(header('Classes', hir + 1))
        obj_tree_to_markdown(cs[0], hir + 2)
        [add(obj_tree_to_markdown(c, hir + 2)) for c in cs]
    if fs:
        add(header('Functions', hir + 1))
        [add(obj_tree_to_markdown(f, hir + 2)) for f in fs]

    return '\n'.join(r)



import re
from collections import OrderedDict as OD
from functools import partial

import structlog
from structlogging import processors as ax_procs

from .common import proc_modules, selected_by_name_processors

try:
    from ax.utils.dynamic_objects import make_staticmethods
    from ax.utils.dynamic_objects import get_mod
except:
    get_mod = None
    make_staticmethods = None


def not_found_processor(logger, level, ev_dict, **kw):
    ev_dict['NOT_FOUND_PROCESSOR'] = kw.pop('proc_str')
    return ev_dict


def build_chain_from_list(processors):
    """get a list of processor names and return a list of processor callables,
    parametrized, if the processor "name" is a dict"""

    # to build the unique name for selective processors:
    # e.g. 'after_add_logger_name_2'
    ctx = {'last': 'beginning', 'nr': 0}

    def processor(p, ctx=ctx):
        if not isinstance(p, (basestring, dict)):
            # already done:
            return p

        cfg = None
        if isinstance(p, dict):
            sub_procs = p.get('selected_by_name')
            # A list of processors run only for some loggers, when the name
            # matches.
            # We build the subchain, indexed by match at global dict
            # 'selected_by_name_processors', so that at logger instantiation
            # we attach the procs to the logger
            # so that the 'selected_processors' processor has an easy time
            # calling them on log events:
            if sub_procs:
                cfg = {
                    'position': 'after_%(last)s_%(nr)s' % ctx,
                    'selector': 'name',
                }
                m = selected_by_name_processors[cfg['position']] = OD()
                for sp in sub_procs:
                    match = sp['match']
                    m[match] = build_chain_from_list(sp['processors'])

                return partial(ax_procs.selected_processors, cfg=cfg)

            # normal processor with some config:
            cfg = p
            p = p.pop('processor')  # crash here, required.

        # todo: custom ones:
        pp = None
        for m in proc_modules:
            pp = getattr(m, p, None)
            if pp:
                break
        if not pp:
            return partial(not_found_processor, proc_str=p)

        # structlog convention: Upper case are classes which can be configured
        if p[0] == p[0].upper():
            return pp(**cfg) if cfg else pp()

        ctx['nr'] += 1
        ctx['last'] = p
        return partial(pp, cfg=cfg) if cfg else pp

    return [processor(p) for p in processors]


class LoggerFactory(structlog.stdlib.LoggerFactory):
    """Wrapper, to attach selective processors"""

    def __init__(self, custom_config, *a, **kw):
        self.custom_config = custom_config
        structlog.stdlib.LoggerFactory.__init__(self, *a, **kw)

    def __call__(self, *args):
        logger = l = structlog.stdlib.LoggerFactory.__call__(self, *args)
        assp = l.ax_structlog_selected_processors = {}
        for pos, chains_by_match in selected_by_name_processors.items():
            assp[pos] = []
            for match in chains_by_match:
                if re.match(match, logger.name):
                    assp[pos].extend(chains_by_match[match])
        return logger


def dict_config(config):
    assert get_mod, 'ax.utils required for structlogging.config'
    d = config.get('structlog')

    # get the list of custom processor modules, where we look for processors:
    while proc_modules:
        proc_modules.pop()
    [proc_modules.append(get_mod(m)) for m in d.get('processor_modules', ())]
    proc_modules.extend([ax_procs, structlog.stdlib, structlog.processors])

    procs = d.get('processors')
    if not procs:
        return

    procs = build_chain_from_list(procs)
    procs.insert(0, structlog.stdlib.filter_by_level)
    procs.append(structlog.stdlib.ProcessorFormatter.wrap_for_formatter)

    structlog.configure(
        processors=procs,
        context_class=dict,
        logger_factory=LoggerFactory(custom_config=d),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def preprocess_dict_config(C):
    """Mutates the Config (C) via batch ops on the dict
    to simplify (re-)config measures
    """
    assert make_staticmethods, 'ax.utils required for preprocess_dict_config'

    def pre_process_dict(C, feat):
        """C the config dict which are mutating"""
        pp = C.get('pre_process_config', {})
        cfg = pp.get(feat)
        if not cfg:
            return
        func = getattr(Features, feat, None)
        if func:
            func(C, cfg)

    class Features:
        def add_handler(C, cfg):
            """add a handler to all matching logs"""

            def add_handler(loggers, match, h):
                [
                    loggers[l]['handlers'].append(h)
                    for l in loggers
                    if re.match(match, l) and not h in loggers[l]['handlers']
                ]

            [add_handler(C['loggers'], match, hdl) for match, hdl in cfg.items()]

        def set_log_dir(C, dir_):
            """prefix all filenames with that dir if """
            for h in C['handlers'].values():
                fn = h.get('filename')
                if fn and not fn.startswith('/'):
                    h['filename'] = '%s/%s' % (dir_, fn)

    make_staticmethods(Features)

    [pre_process_dict(C, f) for f in ['add_handler', 'set_log_dir']]



"""
Taken from axc2 pretty much as is
"""
import collections
import json

import toml

# colors from the os theme for python:
import os
import re
import shutil
import string
import subprocess
import sys
import time
import types
from copy import deepcopy
from fnmatch import fnmatch

# from jsondiff import diff as js_diff # TODO
from functools import partial
from pprint import pformat
from threading import current_thread
from absl import flags
from theming.formatting.markdown import indent, deindent  # noqa

reverse_dict = lambda m: {v: k for k, v in m.items()}


def parse_kw_str(kws, header_kws=None, try_json=True):
    """for kw via cli"""
    header_kws = {} if header_kws is None else header_kws
    if try_json:
        if kws and kws[0] in ('{', '['):
            try:
                return json.loads(kws)
            except:
                pass
    if ', ' in kws:
        raise Exception('No comma allowed')
    kw = {}
    parts = kws.split()
    kw.update(
        {
            p[0]: cast(p[1])
            for p in [(k if '=' in k else k + '=true').split('=') for k in parts]
        }
    )
    kw = {k: header_kws.get(v, v) for k, v in kw.items()}
    return kw


def local_ips(c=[0]):
    i = c[0]
    if i:
        return i
    ips = subprocess.check_output(['hostname', '-I'])
    c[0] = ips.decode('utf-8').strip().split()
    return c[0]


def hostname(c=[0]):
    i = c[0]
    if i:
        return i
    c[0] = socket.gethostname()
    return c[0]


def gitcmd(dir, cmd='git rev-parse --verify HEAD'):
    cmds = [cmd, 'git config --get remote.origin.url']
    t = 'cd "%s" && %s'
    while len(dir) > 2:
        dir = dirname(dir)
        if exists(dir + '/.git/objects'):
            res = [os.popen(t % (dir, cmd)).read().strip() for cmd in cmds]
            return (
                dict(dir=dir, name=os.path.basename(dir), cmd=res[0], url=res[1])
                if res[0]
                else None
            )


def flatten(d, sep='_'):
    import collections

    obj = collections.OrderedDict()

    def recurse(t, parent_key=''):

        if isinstance(t, list):
            for i in range(len(t)):
                recurse(t[i], parent_key + sep + str(i) if parent_key else str(i))
        elif isinstance(t, dict):
            for k, v in t.items():
                recurse(v, parent_key + sep + k if parent_key else k)
        else:
            obj[parent_key] = t

    recurse(d)

    return obj


def cast(v):
    if v and v[0] in ('{', '['):
        return json.loads(v)
    try:
        return int(v)
    except:
        try:
            return float(v)
        except:
            return {'true': True, 'false': False}.get(v, v)


# --------------------------------------------------------------------------------- tty
have_tty = lambda: sys.stdin.isatty() and sys.stdout.isatty()

try:
    # Setting the default for the cli - flag, i.e. this rules w/o the flag:
    term_fixed_width_env = int(os.environ.get('term_width'))
except:
    term_fixed_width_env = 0
try:
    # Setting the default for the cli - flag, i.e. this rules w/o the flag:
    term_fixed_height_env = int(os.environ.get('term_height'))
except:
    term_fixed_height_env = 0


def walk_dir(directory, crit=None):
    crit = (lambda *a: True) if crit is None else crit
    files = []
    j = os.path.join
    for (dirpath, dirnames, filenames) in os.walk(directory):
        files += [j(dirpath, file) for file in filenames if crit(dirpath, file)]
    return files


from contextlib import contextmanager


class Pytest:
    """Helpers specifically for pytest"""

    this_test = lambda: os.environ['PYTEST_CURRENT_TEST']
    cur_flags = {}
    had_flags = {}

    @contextmanager
    def set_flags(**testflags):
        """intended to be run *after* Pytest.init"""
        Pytest.cur_flags.update(testflags)
        f = testflags.items()
        [into(Pytest.had_flags, k, getattr(flags.FLAGS, k)) for k in testflags]
        [setattr(flags.FLAGS, k, v) for k, v in f]
        yield StopIteration
        [setattr(flags.FLAGS, k, v) for k, v in Pytest.had_flags.items()]
        Pytest.cur_flags.clear()
        Pytest.had_flags.clear()

    def parse_test_infos(into=None):
        m = {} if into is None else into
        t = Pytest.this_test()
        f = t.split('::')
        m['class'] = f[1]
        m['test'] = f[-1].split('(')[0].strip()  # testname
        m['file'] = _ = f[0]
        m['fn_mod'] = _ = _.rsplit('/', 1)[-1].rsplit('.py', 1)[0]
        mod = sys.modules[_]
        m['path'] = mod.__file__
        return m

    this_test_func_name = lambda: Pytest.parse_test_infos()['test']

    def set_sys_argv(done=[0]):
        """
        remove any pytest cli args, like -xs, sind that would break FLGs:

        coverage:pytest.py, else pytest. and yes, its shit.
        TODO: create a conf + fixture, avoiding this:
        """
        is_pytest_run = 'pytest' in sys.argv[0].rsplit('/', 1)[-1]

        if done[0] or not is_pytest_run:
            return
        done[0] = True

        while len(sys.argv) > 1:
            sys.argv.pop()

        # strangly we lack our export env, e.g. port-offset in pytest runs since 2020/12
        # happens only if we start via make test, not when started from poetry shell.
        # not sure of a bug of us or a feature of pytest

        # Actual workaround, but deemed anyway better,is to read the flags from here:
        f = '%s/config/pytest.flags' % project.root()
        if os.path.exists(f):
            sys.argv.append('--flagfile=%s' % f)
        # 'environ_flags': True,
        # 'port_offset': os.environ.get('port_offset', 0),
        # 'make_autodocs': os.environ.get('make_autodocs', False),
        # 'log_level': 10,
        # 'log_time_fmt': 'dt',
        # 'log_add_thread_name': True,
        # 'log_dev_fmt_coljson': 'pipes,cfg,payload',
        # 'plot_depth': 3,
        # 'plot_id_short_chars': 8,
        # }

        dflts = {}
        if sys.stdout.isatty():
            dflts['plot_before_build'] = True

        for k, v in dflts.items():
            v = envget(k, v)
            print('pytest EnvFlag: %s = %s' % (k, v))
            if v:
                env[k] = str(v)

    def init():
        Pytest.set_sys_argv()
        FLG(sys.argv)  # parsing the flags


import fcntl, termios, struct


def terminal_size():
    # try piped left and/or right, if both piped then return 80,25:
    # since we want to print we try first stdout:
    for fd in 1, 0:
        try:
            h, w, hp, wp = struct.unpack(
                'HHHH',
                fcntl.ioctl(fd, termios.TIOCGWINSZ, struct.pack('HHHH', 0, 0, 0, 0)),
            )
            return (w or 80), h  # go really sure its no 0
        except:
            pass
    return 80, 25


def termwidth():
    return termsize()[0]


def termsize():
    try:
        w, h = FLG.term_fixed_width, FLG.term_fixed_height
    except:
        # not yet parsed?
        w, h = 0, 0
    if w and h:
        return w, h
    return terminal_size()


# -------------------------------------------------------------------------- data utils
def to_list(o):
    o = [] if o is None else o
    t = type(o)
    return o if t == list else list(o) if t == tuple else [o]


def matched(d, match, prefix=()):
    """Returns parts of data where keys match match string - mainly for debugging (P)"""
    if isinstance(d, (tuple, list)):
        v = [matched(l, match) for l in d]
        if not any([k for k in v if k is not None]):
            return None
        # remove structs wich don't match but leave data:
        r, rd = [], []
        for o in v:
            if isinstance(o, (dict, list, tuple)):
                rd.append(o)
            r.append(o)
        return rd  # best
    if not isinstance(d, dict):
        return
    match = '*%s*' % match
    r = {}
    for k, v in d.items():
        k = str(k)
        np = prefix + (k,)
        if fnmatch(k, match):
            r['.'.join(np)] = v
        vv = matched(v, match, prefix=np)
        if vv:
            if isinstance(v, dict):
                if vv:
                    r.update(vv)
            elif isinstance(v, (tuple, list)):
                r['.'.join(np)] = vv

    return r


def headerize(dicts_list, mark=None):
    '''mark e.g. "dicts"'''
    l = dicts_list
    if not isinstance(l, list) or not l or not isinstance(l[0], dict):
        return l

    r = [list(l[0].keys())]
    r.extend([[m.get(h) for h in r[0]] for m in l])
    if mark:
        r = {mark: r}
    return r


def recurse_data(data, key_callbacks):
    if isinstance(data, dict):
        for k, v in data.items():
            cb = key_callbacks.get(k)
            if cb:
                data[k] = v = cb(v)
            if isinstance(v, (list, dict, tuple)):
                data[k] = v = recurse_data(v, key_callbacks)
    elif isinstance(data, list):
        data = [recurse_data(d, key_callbacks) for d in data]
    elif isinstance(data, tuple):
        data = tuple([recurse_data(d, key_callbacks) for d in data])
    return data


termcols = shutil.get_terminal_size((80, 20)).columns


def P(data, depth=None, match=None, headered=None, out=True, **kw):
    """
    Pretty Printer for large Dicts. In globals(). For debugging.
    P(data, 0, 'wifi') -> deep subtree match

    P(data, depth=2, match=None, [headered|h]='all')

    headered='all' -> all values for key 'all', if a list of dicts, will be printed as lists with headers
    h='all': alias for headered


    kw: pformat kws: indent=1, width=80, depth=None, stream=None, *, compact=False, sort_dicts=True
    """
    # P(data, 'match') also:
    kw['width'] = kw.get('width') or termcols
    headered = kw.pop('h', headered)
    if headered:
        data = deepcopy(data)
        kw['compact'] = kw.get('compact', True)

    if isinstance(depth, str):
        match = depth
        depth = None
    depth = depth or None  # 0 is None
    if match:
        data = matched(data, match)
    h = to_list(headered)
    if h:
        h = {k: partial(headerize, mark='@DL') for k in h}
        data = recurse_data(data, h)
    p = pformat(data, depth=depth, **kw)
    if out:
        print(p)
    else:
        return p


class DictTree(dict):
    __is_tree__ = True

    @property
    def __class__(self):
        # isinstance is dict:
        return dict

    def __getattr__(self, name):
        v = self[name]
        if isinstance(v, dict):
            return v if hasattr(v, '__is_tree__') else DictTree(v)
        return v


class DictTreeMatched(dict):
    __is_tree__ = True

    @property
    def __class__(self):
        # isinstance is dict:
        return dict

    def __getattr__(self, name, nil=b'\x04'):
        v = self.get(name, nil)
        if v == nil:
            for k in self.keys():
                if k.startswith(name):
                    return getattr(self, k)
            self[name]

        if isinstance(v, dict):
            return v if hasattr(v, '__is_tree__') else DictTreeMatched(v)
        return v


# debugging tool: P('data', 1) in pdb sessions
if isinstance(__builtins__, dict):
    __builtins__['P'] = P
    __builtins__['DT'] = DictTree
    __builtins__['T'] = DictTreeMatched
else:
    __builtins__.P = P
    __builtins__.DT = DictTree
    __builtins__.T = DictTreeMatched


def get_deep(key, data, sep='.', create=False, dflt=None):
    """
    Client can, via the dflt. decide how to handle problems, like collisions with values
    """
    # key maybe already path tuple - or string with sep as seperator:
    if not key:
        return data
    d = data
    parts = key.split(sep) if isinstance(key, str) else list(key)
    while parts:
        part = parts.pop(0)
        try:
            data = data[part]
        except TypeError as ex:
            # a list was in the original data:
            try:
                data = data[int(part)]
            except:
                # this happens when we are already at a value, like an int
                # client wants to go deeper, not possible, but we can't delete the int
                # -> leave the client to decide:
                if dflt is None:
                    raise KeyError(key)
                return dflt
        except KeyError as ex:
            if not create:
                if dflt is None:
                    raise
                return dflt
            data = data.setdefault(part, {})

    return data


def deep_update(orig_dict, new_dict, maptyp=collections.abc.Mapping):
    for key, val in new_dict.items():
        if isinstance(val, maptyp):
            tmp = deep_update(orig_dict.get(key, {}), val)
            orig_dict[key] = tmp
        elif isinstance(val, list):
            orig_dict[key] = orig_dict.get(key, []) + val
        else:
            orig_dict[key] = new_dict[key]
    return orig_dict


def in_gevent():
    try:
        import gevent

        return gevent.sleep == time.sleep
    except:
        return False


has_tty = lambda: sys.stdin.isatty() and sys.stdout.isatty()


def confirm(msg):
    if not has_tty():
        raise Exception('Require confirmation for: %s' % msg)
    print(msg)
    r = input('Confirmed [y|N]? ')
    if r.lower() != 'y':
        from devapp.app import app

        app.die('Unconfirmed')


now = lambda: int(time.time() * 1000)
is_ = isinstance


def jdiff(d1, d2):
    try:
        # Got exception when we have L(list) in test_share.py
        return jsondiff(d1, d2, marshal=True)
    except Exception as ex:
        # TODO: convert tuples recursively into strings
        try:
            d1 = json.loads(json.dumps(d1, default=str))
            d2 = json.loads(json.dumps(d2, default=str))
            return jsondiff(d1, d2, marshal=True)
        except Exception as ex:
            return {'err': 'cannot diff', 'exc': str(ex)}


def dict_merge(source, destination):
    """dest wins"""

    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            dict_merge(value, node)
        else:
            destination[key] = value

    return destination


tn = lambda: current_thread().name
pass_ = lambda *a, **kw: None
exists = os.path.exists
dirname = os.path.dirname
abspath = os.path.abspath
env = os.environ
envget = os.environ.get
# kw: rscs may later provide e.g. their name, allowing port offset specific for them:
def offset_port(port, **kw):
    return int(port) + FLG.port_offset


import socket


def wait_for_port(port, host='127.0.0.1', timeout=5.0, log_err=True):
    """Wait until a port starts accepting TCP connections.
    Raises:
        TimeoutError: The port isn't accepting connection after time specified in `timeout`.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                return True
        except OSError as ex:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                if not log_err:
                    return
                try:
                    from devapp.app import app

                    return app.error('Timeout', host=host, port=port, timeout=timeout)
                except:
                    print('timeout awaiting port %s' % port)


def repl_dollar_var_with_env_val(
    s, die_on_fail=True, ask_on_fail=False, get_vals=False
):
    """
    s = foo_$bar.or${baz}  - the first form searched until first non Letter char
    Will be replaced by value of environ for $bar and $baz
    """
    if not '$' in s:
        if not get_vals:
            return s
        return s, {}

    tty, vals = have_tty(), {}

    def go(s, ask_on_fail, tty=tty, vals=vals):
        s += ' '
        missing, have, Letters, nil = {}, [], string.ascii_letters + '_', '\x1b'
        parts = s.split('$')
        n = parts[0]
        for p in parts[1:]:
            if not p:
                raise
            if p[0] == '{':
                if not '}' in p:
                    raise
                have.append(p.split('}', 1)[0][1:])
            else:
                S = ''
                for c in p:
                    S += c
                    if not c in Letters:
                        break
                have.append(S[:-1])
                s = s.replace('$%s' % S, '${%s}%s' % (have[-1], c))
        for k in have:
            app = import_app()
            if not k in env and ask_on_fail and tty:
                app.warn('Missing value in environ', required=k, within=s)
                v = input('Enter value for "$%s": ' % k)
            else:
                v = env[k]
            s = s.replace('${%s}' % k, v)
            vals[k] = v
        return s[:-1]

    r = s
    try:
        r = go(s, ask_on_fail)
        if '$' in r:
            raise
    except Exception as ex:
        if die_on_fail or ask_on_fail:
            app = import_app()
            app.die('Not defined in environ', var=s)
    if get_vals:
        return r, vals
    return r


def import_app():
    from devapp.app import app

    return app


# here = dir_of(__file__)
def dir_of(fn, up=0):
    d = abspath(dirname(fn))
    return dir_of(d, up - 1) if up else d


def to_url(shorthand):
    s = shorthand
    if not s.startswith('http'):
        s = 'http://' + s
    schema, rest = s.split('://', 1)
    if not '/' in rest:
        rest = rest + '/'
    hp, path = rest.split('/', 1)
    if hp.startswith(':'):
        hp = '127.0.0.1' + hp
    elif hp.isdigit():
        hp = '127.0.0.1:%s' % hp
    hp = hp.replace('*', '0.0.0.0')
    return schema + '://' + hp + '/' + path


def host_port(shorthand):
    u = to_url(shorthand).split('//', 1)[1]
    h, p = u.split(':')
    return h, int(p.split('/', 1)[0])


def write_config_cls_key(k, v):
    fn = env['DA_DIR'] + '/config.cls'
    s = read_file(fn).splitlines()
    s = [l for l in s if not l.startswith('%s=' % k)]
    s.append('%s="%s"' % (k, v))
    s.sort()
    write_file(fn, '\n'.join(s) + '\n')


archives = {'.tar.gz': 'tar xfvz "%(frm)s"', '.tar': 'tar xfv "%(frm)s"'}


def download_file(url, local_filename, auto_extract=True):
    """auto extract works for exe types like hugo.tar.gz -> contains in same dir the hugo binary.
    if the .tar.gz is containing a bin dir then the user should rather not say type
    is 'exe' but type is archive or so
    """
    import requests
    from devapp.app import app

    app.info('Downloading', url=url, to=local_filename)
    r = requests.get(url, stream=True)
    arch, fn = None, local_filename
    if auto_extract:
        for k in archives:
            if url.endswith(k):
                arch = k
                fn = local_filename + k
                break

    with open(fn, 'wb') as f:
        shutil.copyfileobj(r.raw, f)
    if arch:
        m = {'dir': fn.rsplit('/', 1)[0], 'cmd': archives[arch], 'frm': fn}
        m['cmd'] = m['cmd'] % m
        cmd = 'cd "%(dir)s" && %(cmd)s' % m
        app.info('Uncharchiving', cmd=cmd)
        os.system(cmd)
    return local_filename


_is_root = [None]  # True, False. None: undefined
_sudo_pw = [None]  # None: undef


def sudo_pw():
    fns = env.get('DA_FILE_SUDO_PASSWORD') or env['HOME'] + '/.sudo_password'
    return env.get('SUDO_PASSWORD') or read_file(fns, '').strip()


def sp_call(*args, as_root='false', get_all=False):
    """run command - as root if 'true' or True"""
    # todo: as_root = 'auto'
    sp = subprocess
    sudo = str(as_root).lower() == 'true'
    if sudo:
        if _is_root[0] == None:
            _is_root[0] = os.system('touch /etc/hosts 2>/dev/null') == 0
        if _is_root[0]:
            sudo = False
    if sudo:
        args = ('sudo', '-S') + tuple(args)
        if _sudo_pw[0] == None:
            _sudo_pw[0] = sudo_pw()
        pw = sp.Popen(('echo', _sudo_pw[0]), stdout=sp.PIPE)
        stdin = pw.stdout
    else:
        stdin = sys.stdin
    proc = sp.Popen(list(args), shell=False, stdin=stdin, stderr=sp.PIPE)
    out, err = proc.communicate()
    if not get_all:
        return err
    return {'exit_status': int(proc.wait()), 'stderr': err, 'stdout': '<term>'}


def exec_file(fn, ctx=None, get_src=None):
    ctx = ctx if ctx is not None else {}
    # allows tracebacks with file pos:
    ctx['__file__'] = fn
    src = read_file(fn)
    exec(compile(src, fn, 'exec'), ctx)
    if get_src:
        ctx['source'] = src
    return ctx


def funcname(f):
    while True:
        n = getattr(f, '__name__', None)
        if n:
            return n
        f = f.func


env = os.environ

clean_env_key = lambda s: ''.join(
    [c for c in s if c in string.digits + string.ascii_letters]
)


def restart_unshared(name):
    """
    Manual user start of daemons.
    Since we assemble filesystems we dont' want those hang around globally,
    but be gone with session end.
    Systemd does this as well.
    must be called at beginning of the process - sys.argv must be original
    """

    def set_linked_python_as_executable():
        """Systemd starts us with sourceing env then ../run/$DA_CLS -m devapp.run start
        sys.argv then: ['$DA_MAX_DIR/../python/devapp/run.py', 'start']
        When a user copies this from the unit file then a unshare is done.
        That calls the start args again after unshare -- ...
        WE have /bin/env python in hashbang - resulting in differences in the ps ax
        -> we replace:
        """
        if sys.argv[0].endswith('/run.py'):
            sys.argv.pop(0)
            for k in 'devapp.run', '-m', '%(DA_DIR)s/run/%(DA_CLS)s' % env:
                sys.argv.insert(0, k)

    n = name
    if sys.argv[-1] == 'unshared':
        sys.argv.pop()
        env['da_unshared'] = n
    if env.get('da_unshared') != n:
        sys.argv.append('unshared')
        print(
            'Restarting unshared. To prevent: export da_unshared="%s"' % n,
            file=sys.stderr,
        )
        c = ['unshare', '-rm', '--']
        set_linked_python_as_executable()
        c.extend(sys.argv)
        sys.exit(subprocess.call(c))


env, envget, exists = os.environ, os.environ.get, os.path.exists
ctx = {}

str8 = partial(str, encoding='utf-8')
is_str = lambda s: isinstance(s, str)

# not matches classes which are callable:
is_func = lambda f: isinstance(
    f, (types.FunctionType, types.BuiltinFunctionType, partial)
)

json_diff = lambda old, new: json.loads(js_diff(old, new, syntax='explicit', dump=True))


def into(d, k, v):
    """the return is important for e.g. rx"""
    d[k] = v
    return d


def setitem(m, k, v):  # for list comprehensions
    m[k] = v


def deep(m, dflt, *pth):
    """Access to props in nested dicts / Creating subdicts by *pth

    Switch to create mode a bit crazy, but had no sig possibilit in PY2:
    # thats a py3 feature I miss, have to fix like that:
    if add is set (m, True) we create the path in m

    Example:
    deep(res, True), [], 'post', 'a', 'b').append(r['fid'])
        creates in res dict: {'post': {'a': {'b': []}}}  (i.e. as list)
        and appends r['fid'] to it in one go
        create because init True was set

    """
    m, add = m if isinstance(m, tuple) else (m, False)
    keys = list(pth)
    while True:
        k = keys.pop(0)
        get = m.setdefault if add else m.get
        v = dflt if not keys else {}
        m = get(k, v)
        if not keys:
            return m


def start_of(s, chars=100):
    """ to not spam output we often do "foo {a': 'b'...."""
    s = str8(s)
    return s[:100] + '...' if len(s) > 100 else ''


cast_list = lambda v, sep=',': (
    [] if v == '[]' else [s.strip() for s in v.split(sep)] if is_str(v) else v
)


dt_precision = '%.2fs'


def dt_human(ts_start, ts_end=None):
    dt = time.time() - ts_start if ts_end is None else ts_end - ts_start
    if dt > 7200:
        return '%.2fh' % (dt / 3600.0)
    return dt_precision % dt


class AllStatic(type):
    'turn all methods of this class into static methods'
    new = None

    def __new__(cls, name, bases, local):
        for k in local:
            if k.startswith('_'):
                continue
            if is_func(local[k]):
                local[k] = staticmethod(local[k])
        # customization hook:
        if cls.new:
            cls.new(cls, name, bases, local)
        return type.__new__(cls, name, bases, local)


class _ctx:
    def __enter__(self, *a, **kw):
        pass

    __exit__ = __enter__


def parse_host(url, no_port=None):
    if not url.startswith('http'):
        raise Exception(url + ' is no url')
    host = '/'.join(url.split('/', 3)[:3])
    if no_port:
        host = ':'.join(host.split(':')[:2])
    return host


class LazyDict(dict):
    def get(self, key, thunk=None):
        return self[key] if key in self else thunk() if callable(thunk) else thunk

    def set_lazy(self, key, thunk=None, *a, **kw):
        return (
            self[key]
            if key in self
            else dict.setdefault(
                self, key, thunk(*a, **kw) if callable(thunk) else thunk
            )
        )


osenv = os.environ.get
HCOLS = {'M': '#8bd124', 'R': '#FF0000', 'L': '#333399', 'I': '#44FFFF'}

shl = 'shell'


def camel_to_snake(string):
    groups = re.findall('([A-z0-9][a-z]*)', string)
    return '_'.join([i.lower() for i in groups])


def url_to_dir(url):
    # url = url.replace('file://', '') # NO, leave
    for ft in '://', '?', ':':
        url = url.replace(ft, '_')
    return url


def html(s, col):
    return '<font color="%s">%s</font>' % (HCOLS[col], s)


def color(s, col, mode=shl):
    if not osenv(col):
        return str(s)
    # for col term
    if mode == 'html':
        return html(s, col)
    else:
        return str(s)


# faster than calling color func:
sm = '%s%%s\x1B[0m' % ('\x1B%s' % osenv('M', '')[2:])
si = '%s%%s\x1B[0m' % ('\x1B%s' % osenv('I', '')[2:])
sl = '%s%%s\x1B[0m' % ('\x1B%s' % osenv('L', '')[2:])
sr = '%s%%s\x1B[0m' % ('\x1B%s' % osenv('R', '')[2:])
sgr = '%s%%s\x1B[0m' % '\x1B[1;38;5;154m'


def M(s, mode=shl):
    if mode == shl:
        return sm % s
    return color(s, 'M', mode)


def I(s, mode=shl):
    if mode == shl:
        return si % s
    return color(s, 'I', mode)


def L(s, mode=shl):
    if mode == shl:
        return sl % s
    return color(s, 'L', mode)


def R(s, mode=shl):
    if mode == shl:
        return sr % s
    return color(s, 'R', mode)


def GR(s):
    return sgr % s


def check_start_env(req_env):
    """ can we run ? """

    def die(msg):
        print(msg)
        sys.exit(1)

    for k in req_env:
        rd = envget(k)
        if not rd:
            die('$%s required' % k)
    return 'Passed requirements check'


def read_file(fn, dflt=None, bytes=-1, strip_comments=False):
    """
    API function.
    read a file - return a default if it does not exist"""
    if not exists(fn):
        if dflt is not None:
            return dflt
        raise Exception(fn, 'does not exist')
    with open(fn) as fd:
        # no idea why but __etc__hostname always contains a linesep at end
        # not present in source => rstrip(), hope this does not break templs
        res = fd.read(bytes)
        res = res if not res.endswith('\n') else res[:-1]
        if strip_comments:
            lines = res.splitlines()
            res = '\n'.join([l for l in lines if not l.startswith('#')])
        return res


def write_file(fn, s, log=0, mkdir=0, chmod=None, mode='w'):
    'API: Write a file. chmod e.g. 0o755 (as octal integer)'
    from devapp.app import app

    fn = os.path.abspath(fn)

    if log > 0:
        app.info('Writing file', fn=fn)

    if isinstance(s, (list, tuple)):
        s = '\n'.join(s)
    if log > 1:
        sep = '\n----------------------\n'
        ps = (
            s
            if not 'key' in fn and not 'assw' in fn and not 'ecret' in fn
            else '<hidden>'
        )
        app.debug('Content', content=sep + ps + sep[:-1])
    e = None
    for i in 1, 2:
        try:
            with open(fn, mode) as fd:
                fd.write(s)
            if chmod:
                if not isinstance(chmod, (list, tuple)):
                    chmod = [int(chmod)]
                for s in chmod:
                    os.chmod(fn, s)
            return fn
        except IOError as ex:
            if mkdir:
                d = os.path.dirname(fn)
                os.makedirs(d)
                continue
            e = ex
        except Exception as ex:
            e = ex
        raise Exception('Could not write file: %s %s' % (fn, e))


def failsafe(meth, *a, **kw):
    """Spotted log errors at aggressively reconnecting clients, running in greenlets
    in the pytest progs. This allows to wrap these methods.
    """
    # spotted log
    try:
        meth(*a, **kw)
    except Exception as ex:
        print('Failed with %s: %s(%s %s)' % (str(ex), meth, str(a), str(kw)))


# ------------------------------------------------------------------------------- flags


FLG = flags.FLAGS


def set_flag(k, v):
    t = type(getattr(FLG, k))
    if t == list:
        v = cast_list(v)

    setattr(FLG, k, t(v))


def set_flag_vals_from_env():
    """allows so ref $foo as default var a flag"""
    # hmm. pytest -> nrtesting -> run_app in greenlet. How to pass flags, e.g. to drawflow
    ef = FLG.environ_flags or os.environ.get(
        'environ_flags'
    )  # if set we check all from environ
    for f in dir(FLG):
        if ef:
            v = os.environ.get(f)
            if v:
                set_flag(f, v)
                continue
        # wanted always from environ by developer - then he gave a default with $:
        v = getattr(FLG, f)
        if isinstance(v, str) and v.startswith('$'):
            setattr(FLG, f, os.environ.get(v[1:], v))


def shorten(key, prefix, maxlen, all_shorts=None, take=1):
    take, sold = 1, None
    while True:
        s = autoshort(key, prefix, maxlen, take)
        if not s in all_shorts:
            return s
        if s == sold:
            raise Exception('Unresolvable collision in autoshort names: %s' % key)
        sold, take, maxlen = s, take + 1, maxlen + 0


def autoshort(key, prefix, maxlen, take=1):
    ml = maxlen - len(prefix)
    parts = key.split('_')
    if prefix and not any([c for c in prefix if not c in parts[0]]):
        parts.pop(0)
    r = prefix + ''.join([parts.pop(0)[0:take].lower() for i in range(ml) if parts])
    return r[:maxlen]


# fmt:off
_ = [
    [type(None),    'boolean'     ] ,
    [str,           'string'      ] ,
    [int,           'integer'     ] ,
    [float,         'float'       ] ,
    [list,          'list'        ] ,
    [bool,          'boolean'     ] ,
    ['multi_string','multi_string'] ,
    ['multi_enum'  ,'multi_enum'  ] ,
]
_flag_makers = dict({k: getattr(flags, 'DEFINE_' + v) for k, v in _})
# fmt:on


def mk_enum(*a, vals, **kw):
    f = flags.DEFINE_multi_enum if isinstance(vals, tuple) else flags.DEFINE_enum
    a = list(a)
    a.insert(2, vals)
    return f(*a, **kw)


def flag_makers(t, m=_flag_makers):
    # the actual absl.flags.DEFINE_integer(..),... methods
    if isinstance(t, (list, tuple)):
        return partial(mk_enum, vals=t)
    else:
        return _flag_makers[t]


all_flag_shorts = {}


def make_flag(c, module, autoshort, default, sub=False, **kw):
    g = getattr
    key = orig_key = c.__name__
    # if key == 'verbose':
    #     breakpoint()  # FIXME BREAKPOINT
    if sub and not sub == 'Actions':
        key = sub + '_' + key
    d = g(c, 'd', default)  # when d is not given, what to do. Dictates the flag type.
    ml = kw.get('short_maxlen', 5)
    s = g(c, 's', None)
    mkw = {'module_name': module, 'short_name': s}
    if s is None and autoshort is not False:
        s = shorten(orig_key, prefix=autoshort, maxlen=ml, all_shorts=all_flag_shorts)
    if s is False:
        s = None
    else:
        all_flag_shorts[s] = key
        # when we parse the cli for action flags we also want to find shortened afs:
        af = action_flags.get(key)
        if af:
            action_flags[s] = af
    mkw['short_name'] = s
    define_flag = flag_makers(g(c, 't', type(d)))
    txt = g(c, 'n', human(key))
    if c.__doc__:
        ls = c.__doc__.replace(c.n, '').splitlines()
        txt += ' Details: %s' % ('\n'.join([l.strip() for l in ls]).strip())
    define_flag(key, d, txt, **mkw)


human = lambda key: ' '.join([k.capitalize() for k in key.split(' ')])


# this allows a module to "steal" the flags class of another module. e.g. lc client:
skip_flag_defines = []


have_flg_cls = set()
action_flags = {}


def have_subs(c):
    for s in [getattr(c, s) for s in dir(c) if not s.startswith('_')]:
        if isinstance(s, type):
            return True


def rm_absl_flags():
    FLG.remove_flag_values(
        ['v',]
    )


def define_flags(Class, sub=False, parent_autoshort=False):
    """
    Pretty classes to flag defs. See nrclient, devapp.app or structlogging.sl how to use
    All config in the top class
    2021/06: Subclassses allowed
    """

    if Class in have_flg_cls:
        from devapp.app import app

        return app.warn('Flag class defined already - skipping', Class=Class)
    have_flg_cls.add(Class)
    # TODO: Register docstrings here (markdown?), supplying even more detailed help
    g = getattr
    module = g(Class, 'module', Class.__module__)
    if module in skip_flag_defines:
        return
    # passed to make_flag as default for default
    default = g(Class, 'default', '')
    autoshort = parent_autoshort = g(Class, 'autoshort', parent_autoshort)
    l = dict(locals())
    cshrt, c_no_shrt = [], []
    for k in [i for i in dir(Class) if not i.startswith('_')]:
        c = g(Class, k)

        if not isinstance(c, type):
            continue
        if sub == 'Actions':
            setattr(c, 'd', False if not g(c, 'd', None) == True else True)
            action_flags[k] = {
                'flg_cls': c,
                'class': Class,
                'key': k,
                'autoshort': g(c, 'autoshort', autoshort),
                'is_default': c.d,
            }

        if not hasattr(c, 'n'):
            if not hasattr(c, 'd'):
                a = g(c, 'autoshort', parent_autoshort)
                if not a and not c.__name__ == 'Actions':
                    pref = c.__name__[0]
                # prbably group but no n, no d is allowed as well
                if have_subs(c):
                    r = define_flags(c, sub=c.__name__, parent_autoshort=a)
                    continue
            # alternative to stating n we allow a multline docstring, where first line
            # is n, rest is details:
            doc = c.__doc__ or ''
            if doc.strip():
                c.n = doc.strip().split('\n', 1)[0]
        # if g(c, 's', None) == 'x':
        #     breakpoint()  # FIXME BREAKPOINT
        cshrt.append(c) if g(c, 's', None) else c_no_shrt.append(c)
    # do the ones with shorts first, to collide only on autoshorts:
    [make_flag(c, **l) for c in cshrt + c_no_shrt]
    # [make_flag(c, **l) for k in [cshrt, c_no_shrt] for c in k]


# define_flags(Flags)  # ours


def FlagsDict(f):
    """
    Flags to Dicts
    allows:
    FLG.foo = 'bar' -> '%(foo)s' % FlagsDict(Flags) = 'bar'
    i.e. Allow values of Flags to reference values of other flags.
    """

    # class F:
    #     __init__ = lambda self, f: setattr(self, 'f', f)
    #     __getitem__ = lambda self, k: getattr(self.f, k)

    # return F(f)
    return FLG.flag_values_dict()


class project:
    """loads the project config

    Main app config put into project.cfg['app']
    An optional tool.lc.app section will be merged into that

    TODO: other formats than pyproject.toml

    """

    config, dir_home, fn_cfg = {}, None, None

    def root(no_fail=False):
        # r = FLG.project_directory or os.environ.get('project_directory')
        # if r:
        #     if not exists(r):
        #         from devapp.app import app

        #         os.makedirs(r)
        #         os.system('cd "%s"; git init' % r)
        #         app.warn('Created project (git) directory', dir=r)
        #     return r
        p = project.dir_home
        if not p:
            p = project.set_project_dir(no_fail=no_fail)
            if p:
                b = p + '/bin:'
                if not b in envget('PATH', ''):
                    env['PATH'] = b + envget('PATH', '')
            project.dir_home = p
        return project.dir_home

    def set_project_dir(dir=None, no_fail=False):
        """
        We search up and fail if we are not within the dir
        this is consistent with poetry (say poetry debug outside the dir ->
        Poetry could not find a pyproject.toml file in / or its parents
        """
        if dir:
            # hard set, e.g. from project initialize:
            project.dir_home = dir
            return

        d = os.getcwd()
        while len(d) > 3:
            for fn in 'pyproject.toml', 'setup.py', '.git':
                if os.path.exists(d + '/' + fn):
                    return d
            d = os.path.dirname(d)
        if no_fail:
            return
        from devapp.app import app

        app.die('could not determine project root dir')

    fn_resources = lambda: project.root() + '/.resources.json'

    fn_config = lambda: project.root() + '/pyproject.toml'

    def read_resources(filename=None):
        from devapp.app import app

        fn = filename or project.fn_resources()
        s = read_file(fn, dflt='')
        s = json.loads(s) if s else {}
        if s:
            app.info('loaded resources', filename=fn)
        project.resources = s
        return s

    def write_resources(rsc, filename=None):
        from devapp.app import app

        fn = filename or project.fn_resources()
        with open(fn, 'w') as fd:
            fd.write(json.dumps(rsc, indent=4))
        app.info('have written resources file', filename=fn)

    def get_present_resource(rsc, _have={}):
        from devapp.app import app

        if rsc in _have:
            return _have[rsc]
        r = project.read_resources()
        r = [i for i in r if i['name'] == rsc]
        if len(r) == 1 and r[0].get('installed'):
            _have[rsc] = r[0]
            return r[0]
        app.warn('Missing resource', resource=rsc)
        _have[rsc] = None

    def get_present_resource_location(rsc):
        r = project.get_present_resource(rsc)
        if r:
            r = r.get('installed')
            return r

    # TODO: understand also poetry and piptools:
    def load_config():
        from devapp.app import app

        fn = project.fn_config()
        if not exists(fn):
            app.die('no config found', fn=fn)

        cfg = toml.load(fn)
        app.info('loaded config', filename=fn)
        c = project.config
        c.update(cfg)
        if not 'project' in c:
            c['project'] = {'urls': {}}
        if 'tool' in c and 'poetry' in c['tool']:
            c['project'].update(c['tool']['poetry'])
            c['project']['urls']['homepage'] = c['project']['homepage']
            c['project']['urls']['repository'] = c['project']['repository']

        t = cfg['tool']
        # breakpoint()  # FIXME BREAKPOINT
        # c['app'] = t['poetry']
        # c['app'].update(t.get('lc', {}).get('app', {}))
        project.fn_cfg = fn
        # app.die('Did not find a pyproject.toml file with badges declared')
        return project.config

    def conf():
        return project.config or project.load_config()

    def name():
        p = project.conf()
        return p['project']['name']

    def urls():
        p = project.conf()['project']
        if 'urls' in p:
            return p['urls']
        urls = 'packagehome', 'discusshome', 'homepage', 'repository'
        return {k: p.get(k, '') for k in urls}

    def homepage():
        return project.urls().get('homepage', 'n.a.')

    def repository():
        return project.urls().get('repository', 'n.a.')

    def packagehome():
        return project.urls().get('packagehome', 'n.a.')

    def version():
        p = project.conf()['project']
        v = p['version']
        if isinstance(v, dict) and v.get('use_scm'):
            from pdm.pep517.scm import get_version_from_scm

            v = get_version_from_scm(project.root())
        return str(v)

    def dependencies():
        d = project.conf()['project']['dependencies']
        if isinstance(d, dict):
            # poetry:
            return d
        # pdm (pep)
        return parse_deps(d)

    def dev_dependencies():
        p = project.conf()['project']
        dd = p.get('dev-dependencies')
        if dd:
            if isinstance(dd, dict):
                # poetry - already dict:
                return dd
            return parse_deps(dd)

        r = [
            l
            for k in project.conf()['tool']['pdm']['dev-dependencies'].values()
            for l in k
        ]
        return parse_deps(r)

    def lock_data():
        fn = []
        for k in 'pdm', 'poetry':
            fn.insert(0, project.root() + '/%s.lock' % k)
            if exists(fn[0]):
                return toml.load(fn[0])
        die('No lock file in root', fn)


def parse_deps(deplist, seps='~<>!= '):
    m = {}
    for dep in deplist:
        h = False
        for s in seps:
            if s in dep:
                l = dep.split(s, 1)
                m[l[0]] = (s + l[1]).strip()
                h = True
                break
        if not h:
            m[dep] = ''
    return m


import re


is_sensitive = lambda key: re.match(FLG.sensitive_data_identifiers, key, re.IGNORECASE)
filter_passwords = lambda dct: {k: v for k, v in dct.items() if not is_sensitive(k)}


# --------------------------------------------------------------------- Not importable
def unavail(missing, req=None, warn=True):
    """We don't need everything always"""
    # TODO:Add a few pip install options
    def f(*a, _miss=missing, _req=req, _warn=warn, **kw):
        print('Function from package "%s" not available in your installation!' % _miss)
        if _req:
            print('Please run: ', _req)
        raise Exception('Cannot continue - missing %s' % _miss)

    if warn:
        print('Cannot import: %s' % missing)
    return f


try:
    from jsondiff import diff as jsondiff
except Exception as ex:
    jsondiff = unavail('jsondiff', req='pip install jsondiff', warn=False)


# ------------------------------------------------------------ decorators


# https://github.com/micheles/decorator/blob/master/docs/documentation.md
# preserve the signature of decorated functions in a consistent way across Python releases

try:
    from decorator import decorate, decorator
except ImportError as ex:
    decorate = decorator = unavail('decorator', req='pip install decorator', warn=False)

try:
    from tabulate import tabulate
except ImportError as ex:
    tabulate = unavail('tabulate', req='pip install tabulate', warn=False)


def _memoize(func, *args, **kw):
    if kw:  # frozenset is used to ensure hashability
        key = args, frozenset(list(kw.items()))
    else:
        key = args
    cache = func.cache  # public cache attribute added by memoize
    if key not in cache:
        cache[key] = func(*args, **kw)
    return cache[key]


def memoize(f):
    """
    A simple memoize implementation. It works by adding a .cache diction
    to the decorated function. The cache will grow indefinitely, so it i
    your responsibility to clear it, if needed.
    """
    f.cache = {}
    return decorate(f, _memoize)


class appflags:
    default = False
    autoshort = 'da'
    uc_dirwatch_flow = [
        'Running build app, monitoring flow.json, triggering plots on change',
        'build --plot_before --dirwatch .:flow',
    ]

    class dirwatch:
        n = 'Spawning helper process, sending reload, on changes in given directory. '
        n += 'The app itself keeps running normally, with full access at stdin,out, err. '
        n += 'You may deliver a match spec, colon seperated.'
        d = ''
        h = 'Format: <dir>[:match[:r]]. r for recursive. match w/o * will be enclosed in *, left and right.'
        s = 'dw'

    class redir_stderr:
        """There are process spawners which mix the streams (emacs call-process)
        Spawns sys.exit(os.system(cmd + ' 2>%s' % FLG.redir_stderr))"""

        n = 'Global redirect of stderr'
        d = ''

    class flat:
        n = 'Flatten output of nested structures'

    class sensitive_data_identifiers:
        n = 'Regexp which helps identify keys carrying sensitive information (for filtering out of logs). Case insensitive matching.'
        d = 'pass.*|.*secret.*'

    class help_call:
        n = 'Short help about a call'
        s = 'hh'

    class help_call_detailed:
        n = 'Detailed help about a call'
        s = 'hhh'

    class environ_flags:
        n = 'Respect environ first, for all flags'

    class port_offset:
        n = 'Set a global offset for any "port" argument of a function'
        d = 0

    class term_fixed_width:
        n = 'Sets terminal width to this. 0: auto.'
        d = term_fixed_width_env

    class term_fixed_height:
        n = 'Sets terminal size to this. 0: auto.'
        d = term_fixed_height_env

    # not needed to run os.system calls anymore:
    # class term_auto_width_recalc_every:
    #     n = 'When term with is auto we run stty size maximum every so many millisecs'
    #     d = 2000

    class help_output_fmt:
        s = False  #  must be long form, checked in absl_color_help
        t = [
            'plain',
            'simple',
            'github',
            'grid',
            'fancy_grid',
            'pipe',
            'orgtbl',
            'jira',
            'presto',
            'pretty',
            'psql',
            'rst',
            'mediawiki',
            'moinmoin',
            'youtrack',
            'html',
            'latex',
            'latex_raw',
            'latex_booktabs',
            'terminal',
            'textile',
        ]
        d = 'terminal'


define_flags(appflags)


if __name__ == '__main__':
    pass
    # r = P(
    #     {'a': [{'all': [{'a': 23, 'b': 23}, {'a': 2322, 'b': 232323}]}]},
    #     out=None,
    #     headered='all',
    # )
    # assert r == "{'a': [{'all': [['a', 'b'], 23, 23, 2322, 232323]}]}"



def cls_with_meta(mc, attrs):
    class _tree_builder_intermediate_(metaclass=mc):
        pass

    for k, v in attrs.items():
        setattr(_tree_builder_intermediate_, k, v)
    return _tree_builder_intermediate_



#!/usr/bin/env python
# TODO: make this a real program, currently just a hack to get tokens
"""
for now just run this with a >> to your tokens store file

"""
import os
import sys
import time

import jwt


def run():
    here = os.path.abspath(os.path.dirname(__file__))
    sys.argv.append('')
    sys.argv.append('')

    secret_file = sys.argv[1]

    why = sys.argv[2]
    if not why:
        print('add why as sys.argv[2]')
        sys.exit(1)
    realms = sys.argv[3]
    if not realms:
        print('add realms as sys.argv[3]')
        sys.exit(1)

    now = int(time.time())
    now5y = now + 86400 * 360 * 5
    message = {
        'iss': 'https://axwifi.com/',
        'sub': 'ax.wifi',
        # 'roles': 'publ_*',
        'roles': sys.argv[4],
        'realms': realms,
        'why': why,
        'iat': now,
        'exp': now5y,
        'max_rate': 10,  # per sec
    }

    # print('\n========== NEW TOKEN==%s==%s====\n' % (time.ctime(), why))
    [print(k, v, file=sys.stderr) for k, v in list(message.items())]
    print(
        str(
            jwt.encode(
                message, open(secret_file).read().strip(), algorithm='HS256'
            ).decode('utf-8')
        )
    )
